# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from subprocess import Popen
from shutil import which
from typing import Any
from collections.abc import Sequence

from xpra.net.compression import Compressed
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.common import FULL_INFO, NotificationID, SizedBuffer
from xpra.os_util import get_machine_id, get_user_uuid, gi_import
from xpra.util.objects import typedict
from xpra.util.str_fn import csv, bytestostr
from xpra.util.env import envint, envbool, first_time
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("audio")

NEW_STREAM_SOUND = envbool("XPRA_NEW_STREAM_SOUND", True)
NEW_STREAM_SOUND_STOP = envint("XPRA_NEW_STREAM_SOUND_STOP", 20)


class FakeSink:
    def __init__(self, codec: str):
        self.codec = codec

    @staticmethod
    def add_data(*args) -> None:
        log("FakeSink.add_data%s ignored", args)

    @staticmethod
    def cleanup(*args) -> None:
        log("FakeSink.cleanup%s ignored", args)


def stop_proc(proc) -> None:
    r = proc.poll()
    log("stop_proc(%s) exit code=%s", proc, r)
    if r is not None:
        # already ended
        return
    try:
        proc.terminate()
    except OSError:
        log("failed to stop subprocess %s", proc)


class AudioMixin(StubSourceMixin):

    PREFIX = "audio"

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        audio = caps.get(AudioMixin.PREFIX)
        if isinstance(audio, dict):
            audio = typedict(audio)
            return audio.boolget("send") or audio.boolget("receive")
        return False

    def __init__(self):
        self.audio_properties: typedict = typedict()
        self.audio_source_plugin = ""
        self.supports_speaker = False
        self.speaker_codecs = []
        self.supports_microphone = False
        self.microphone_codecs = []
        self.restart_speaker_args = ()
        self.audio_suspended = False

    def init_from(self, _protocol, server) -> None:
        self.audio_properties = typedict(server.audio_properties)
        self.audio_source_plugin = server.audio_source_plugin
        self.supports_speaker = server.supports_speaker
        self.supports_microphone = server.supports_microphone
        self.speaker_codecs = server.speaker_codecs
        self.microphone_codecs = server.microphone_codecs

    def init_state(self) -> None:
        self.wants_audio = True
        self.audio_source_sequence = 0
        self.audio_source: Any = None
        self.audio_sink: Any = None
        self.pulseaudio_id = ""
        self.pulseaudio_cookie_hash = ""
        self.pulseaudio_server = ""
        self.audio_decoders: Sequence[str] = ()
        self.audio_encoders: Sequence[str] = ()
        self.audio_receive = False
        self.audio_send = False
        self.audio_fade_timer = 0
        self.new_stream_timers: dict[Popen, int] = {}

    def cleanup(self) -> None:
        log("%s.cleanup()", self)
        self.cancel_audio_fade_timer()
        self.stop_sending_audio()
        self.stop_receiving_audio()
        self.stop_new_stream_notifications()
        self.init_state()

    def stop_new_stream_notifications(self) -> None:
        timers = self.new_stream_timers.copy()
        self.new_stream_timers = {}
        for proc, timer in timers.items():
            timer = self.new_stream_timers.pop(proc, 0)
            if timer:
                GLib.source_remove(timer)
            stop_proc(proc)

    def parse_client_caps(self, c: typedict) -> None:
        audio = typedict(c.dictget(AudioMixin.PREFIX) or {})
        self.wants_audio = "audio" in c.strtupleget("wants") or audio.boolget("send") or audio.boolget("receive")
        if audio:
            self.pulseaudio_id = audio.strget("pulseaudio.id")
            self.pulseaudio_cookie_hash = audio.strget("pulseaudio.cookie-hash")
            self.pulseaudio_server = audio.strget("pulseaudio.server")
            self.audio_decoders = audio.strtupleget("decoders", ())
            self.audio_encoders = audio.strtupleget("encoders", ())
            self.audio_receive = audio.boolget("receive")
            self.audio_send = audio.boolget("send")
        log("pulseaudio id=%s, cookie-hash=%s, server=%s, audio decoders=%s, audio encoders=%s, receive=%s, send=%s",
            self.pulseaudio_id, self.pulseaudio_cookie_hash, self.pulseaudio_server,
            self.audio_decoders, self.audio_encoders, self.audio_receive, self.audio_send)

    def get_caps(self) -> dict[str, Any]:
        if not self.wants_audio or not self.audio_properties:
            return {}
        audio_props = dict(self.audio_properties)
        if FULL_INFO < 2:
            # only expose these specific keys:
            audio_props = {k: v for k, v in audio_props.items() if k in (
                "muxers", "demuxers",
            )}
        audio_props.update({
            "codec-full-names": True,
            "encoders": self.speaker_codecs,
            "decoders": self.microphone_codecs,
            "send": self.supports_speaker and len(self.speaker_codecs) > 0,
            "receive": self.supports_microphone and len(self.microphone_codecs) > 0,
        })
        return {"audio": audio_props}

    def audio_loop_check(self, mode: str = "speaker") -> bool:
        log("audio_loop_check(%s)", mode)
        # pylint: disable=import-outside-toplevel
        from xpra.audio.gstreamer_util import ALLOW_SOUND_LOOP, loop_warning_messages
        if ALLOW_SOUND_LOOP:
            return True
        machine_id = get_machine_id()
        uuid = get_user_uuid()
        # these attributes belong in a different mixin,
        # so we can't assume that they exist:
        client_machine_id = getattr(self, "machine_id", None)
        client_uuid = getattr(self, "uuid", None)
        log("audio_loop_check(%s) machine_id=%s client machine_id=%s, uuid=%s, client uuid=%s",
            mode, machine_id, client_machine_id, uuid, client_uuid)
        if client_machine_id:
            if client_machine_id != machine_id:
                # not the same machine, so OK
                return True
            if client_uuid != uuid:
                # different user, assume different pulseaudio server
                return True
        # check pulseaudio id if we have it
        padict = typedict(self.audio_properties.get("pulseaudio", {}))
        pulseaudio_id = padict.strget("id")
        pulseaudio_cookie_hash = padict.strget("cookie-hash")
        log("audio_loop_check(%s) pulseaudio id=%s, client pulseaudio id=%s",
            mode, pulseaudio_id, self.pulseaudio_id)
        log("audio_loop_check(%s) pulseaudio cookie hash=%s, client pulseaudio cookie hash=%s",
            mode, pulseaudio_cookie_hash, self.pulseaudio_cookie_hash)
        if pulseaudio_id and self.pulseaudio_id:
            if self.pulseaudio_id != pulseaudio_id:
                return True
        elif pulseaudio_cookie_hash and self.pulseaudio_cookie_hash:
            if self.pulseaudio_cookie_hash != pulseaudio_cookie_hash:
                return True
        else:
            # no cookie or id, so probably not a pulseaudio setup,
            # hope for the best:
            return True
        msgs = loop_warning_messages(mode)
        summary = msgs[0]
        body = "\n".join(msgs[1:])
        self.may_notify(NotificationID.AUDIO, summary, body, icon_name=mode)
        log.warn("Warning: %s", summary)
        for x in msgs[1:]:
            log.warn(" %s", x)
        return False

    def start_sending_audio(self, codec: str = "", volume: float = 1.0):
        log("start_sending_audio(%s)", codec)
        ss = None
        if getattr(self, "suspended", False):
            log.warn("Warning: not starting audio whilst in suspended state")
            return None
        if not self.supports_speaker:
            log.error("Error sending audio: support not enabled on the server")
            return None
        if self.audio_source:
            log.error("Error sending audio: forwarding already in progress")
            return None
        if not self.audio_receive:
            log.error("Error sending audio: support is not enabled on the client")
            return None
        if not codec:
            codecs = [x for x in self.audio_decoders if x in self.speaker_codecs]
            if not codecs:
                log.error("Error sending audio: no codecs in common")
                return None
            codec = codecs[0]
        elif codec not in self.speaker_codecs:
            log.warn("Warning: invalid codec specified: %s", codec)
            return None
        elif codec not in self.audio_decoders:
            log.warn("Error sending audio: invalid codec '%s'", codec)
            log.warn(" is not in the list of decoders supported by the client: %s", csv(self.audio_decoders))
            return None
        if not self.audio_loop_check("speaker"):
            return None
        try:
            from xpra.audio.wrapper import start_sending_audio  # pylint: disable=import-outside-toplevel
            plugins = self.audio_properties.strtupleget("sources")
            ss = start_sending_audio(plugins, self.audio_source_plugin,
                                     "", codec, volume, True, [codec],
                                     self.pulseaudio_server, self.pulseaudio_id)
            self.audio_source = ss
            log("start_sending_audio() audio source=%s", ss)
            if not ss:
                return None
            ss.sequence = self.audio_source_sequence
            ss.connect("new-buffer", self.new_audio_buffer)
            ss.connect("new-stream", self.new_stream)
            ss.connect("info", self.audio_source_info)
            ss.connect("exit", self.audio_source_exit)
            ss.connect("error", self.audio_source_error)
            ss.start()
            self.restart_speaker_args = (codec, volume)
            return ss
        except Exception as e:
            log.error("Error setting up audio: %s", e, exc_info=True)
            self.stop_sending_audio()
            ss = None
            return None
        finally:
            if ss is None:
                # tell the client we're not sending anything:
                self.send_eos(codec)

    def audio_source_error(self, source, message: str) -> None:
        # this should be printed to stderr by the audio process already
        if source == self.audio_source:
            log("audio capture error: %s", message)

    def audio_source_exit(self, source, *args) -> None:
        log("audio_source_exit(%s, %s)", source, args)
        if source == self.audio_source:
            self.stop_sending_audio()

    @staticmethod
    def audio_source_info(source, info: dict) -> None:
        log("audio_source_info(%s, %s)", source, info)

    def stop_sending_audio(self) -> None:
        ss = self.audio_source
        log("stop_sending_audio() audio_source=%s", ss)
        if ss:
            self.audio_source = None
            self.send_eos(ss.codec, ss.sequence)
            self.audio_source_sequence += 1
            ss.cleanup()
        self.call_update_av_sync_delay()

    def send_eos(self, codec: str, sequence: int = 0) -> None:
        log("send_eos(%s, %s)", codec, sequence)
        # tell the client this is the end:
        self.send_more("sound-data", codec, b"",
                       {
                           "end-of-stream": True,
                           "sequence": sequence,
                       })

    def new_stream_sound(self) -> None:
        if not NEW_STREAM_SOUND:
            return
        from xpra.platform.paths import get_resources_dir  # pylint: disable=import-outside-toplevel
        sample = os.path.abspath(os.path.normpath(os.path.join(get_resources_dir(), "bell.wav")))
        log(f"new_stream_sound() sample={sample}, exists={os.path.exists(sample)}")
        if not os.path.exists(sample):
            return
        gst_launch = os.path.abspath(os.path.normpath(which("gst-launch-1.0") or "gst-launch-1.0"))
        cmd = [
            gst_launch, "-q",
            "filesrc", f"location={sample}",
            "!", "decodebin",
            "!", "audioconvert",
            "!", "autoaudiosink",
        ]
        cmd_str = " ".join(cmd)
        try:
            proc = Popen(cmd)  # pylint: disable=consider-using-with
            log(f"Popen({cmd_str})={proc}")
            from xpra.util.child_reaper import getChildReaper  # pylint: disable=import-outside-toplevel
            getChildReaper().add_process(proc, "new-stream-sound", cmd, ignore=True, forget=True)

            def stop_new_stream_notification() -> None:
                if self.new_stream_timers.pop(proc, None):
                    stop_proc(proc)

            timer = GLib.timeout_add(NEW_STREAM_SOUND_STOP * 1000, stop_new_stream_notification)
            self.new_stream_timers[proc] = timer
        except Exception as e:
            log("new_stream_sound() error playing new stream sound", exc_info=True)
            log.error("Error playing new-stream sound")
            log.error(f" using: {cmd_str}:")
            log.estr(e)

    def new_stream(self, audio_source, codec: str) -> bool:
        log("new_stream(%s, %s)", audio_source, codec)
        self.new_stream_sound()
        if self.audio_source != audio_source:
            log("dropping new-stream signal (current source=%s, signal source=%s)", self.audio_source, audio_source)
            return False
        codec = codec or audio_source.codec
        audio_source.codec = codec
        # tell the client this is the start:
        self.send("sound-data", codec, b"", {
            "start-of-stream": True,
            "codec": codec,
            "sequence": audio_source.sequence,
        })
        self.call_update_av_sync_delay()
        # run it again after 10 seconds,
        # by that point the source info will actually be populated:
        GLib.timeout_add(10 * 1000, self.call_update_av_sync_delay)
        return False

    def call_update_av_sync_delay(self) -> None:
        # loose coupling with avsync mixin:
        update_av_sync = getattr(self, "update_av_sync_delay_total", None)
        log("call_update_av_sync_delay update_av_sync=%s", update_av_sync)
        if callable(update_av_sync):
            update_av_sync()  # pylint: disable=not-callable

    def new_audio_buffer(self, audio_source, data: bytes,
                         metadata: dict, packet_metadata: Sequence[SizedBuffer]) -> None:
        log("new_audio_buffer(%s, %s, %s, %s) info=%s",
            audio_source, len(data or []), metadata, [len(x) for x in packet_metadata], audio_source.info)
        if self.audio_source != audio_source or self.is_closed():
            log("audio buffer dropped: from old source or closed")
            return
        if audio_source.sequence < self.audio_source_sequence:
            log("audio buffer dropped: old sequence number: %s (current is %s)",
                audio_source.sequence, self.audio_source_sequence)
            return
        self.send_audio_data(audio_source, data, metadata, packet_metadata)

    def send_audio_data(self, audio_source, data: bytes, metadata: dict,
                        packet_metadata: Sequence[SizedBuffer]) -> None:
        # tag the packet metadata as already compressed:
        pmetadata = Compressed("packet metadata", packet_metadata)
        packet_data = [audio_source.codec, Compressed(audio_source.codec, data), metadata, pmetadata]
        sequence = audio_source.sequence
        if sequence >= 0:
            metadata["sequence"] = sequence
        self.send("sound-data", *packet_data, synchronous=False, will_have_more=True)

    def stop_receiving_audio(self) -> None:
        ss = self.audio_sink
        log("stop_receiving_audio() audio_sink=%s", ss)
        if ss:
            self.audio_sink = None
            ss.cleanup()

    ##########################################################################
    # audio control commands:
    def audio_control(self, action: str, *args):
        fn = "audio_control_" + action.replace("-", "_")
        method = getattr(self, fn, None)
        log(f"audio_control({action}, {args}) {self}.{fn}={method}")
        if not method:
            msg = f"unknown audio action {action!r}"
            if first_time(f"unknown-{method}"):
                log.error("Error: %s", msg)
            return msg
        return method(*args)  # pylint: disable=not-callable

    def audio_control_stop(self, sequence_str="") -> str:
        if sequence_str:
            try:
                sequence = int(sequence_str)
            except ValueError:
                msg = f"audio sequence number {sequence_str!r} is invalid"
                log.warn(msg)
                return msg
            if sequence != self.audio_source_sequence:
                log.warn("Warning: audio sequence mismatch: %i vs %i",
                         sequence, self.audio_source_sequence)
                return "not stopped"
            log("stop: sequence number matches")
        self.stop_sending_audio()
        return "stopped"

    def audio_control_fadein(self, codec: str = "", delay_str="") -> str:
        self.do_audio_control_start(0.0, codec)
        delay = 1000
        if delay_str:
            delay = max(1, min(10 * 1000, int(delay_str)))
        step = 1.0 / (delay / 100.0)
        log("audio_control fadein delay=%s, step=%1.f", delay, step)

        def fadein() -> bool:
            ss = self.audio_source
            if not ss:
                return False
            volume = ss.get_volume()
            log("fadein() volume=%.1f", volume)
            if volume < 1.0:
                volume = min(1.0, volume + step)
                ss.set_volume(volume)
            return volume < 1.0

        self.cancel_audio_fade_timer()
        self.audio_fade_timer = GLib.timeout_add(100, fadein)
        return "fadein started"

    def audio_control_start(self, codec: str = "") -> str:
        self.do_audio_control_start(1.0, codec)
        return f"requested {codec} audio"

    def do_audio_control_start(self, volume: float, codec: str) -> str:
        codec = bytestostr(codec)
        log("do_audio_control_start(%s, %s)", volume, codec)
        if not self.start_sending_audio(codec, volume):
            return "failed to start audio"
        msg = "audio started"
        if codec:
            msg += f" using codec {codec}"
        return msg

    def audio_control_fadeout(self, delay_str="") -> str:
        assert self.audio_source, "no active audio capture"
        delay = 1000
        if delay_str:
            delay = max(1, min(10 * 1000, int(delay_str)))
        step = 1.0 / (delay / 100.0)
        log("audio_control fadeout delay=%s, step=%1.f", delay, step)

        def fadeout() -> bool:
            ss = self.audio_source
            if not ss:
                return False
            volume = ss.get_volume()
            log("fadeout() volume=%.1f", volume)
            if volume > 0:
                ss.set_volume(max(0, volume - step))
                return True
            self.stop_sending_audio()
            return False

        self.cancel_audio_fade_timer()
        self.audio_fade_timer = GLib.timeout_add(100, fadeout)
        return "fadeout started"

    def audio_control_new_sequence(self, seq_str) -> str:
        self.audio_source_sequence = int(seq_str)
        return f"new sequence is {self.audio_source_sequence}"

    def cancel_audio_fade_timer(self) -> None:
        sft = self.audio_fade_timer
        if sft:
            self.audio_fade_timer = 0
            GLib.source_remove(sft)

    def audio_data(self, codec: str, data: bytes, metadata: dict, packet_metadata: Sequence[SizedBuffer] = ()) -> None:
        log("audio_data(%s, %s, %s, %s) audio sink=%s",
            codec, len(data or []), metadata, packet_metadata, self.audio_sink)
        if self.is_closed():
            return
        if self.audio_sink is not None and codec != self.audio_sink.codec:
            log.info("audio codec changed from %s to %s", self.audio_sink.codec, codec)
            self.audio_sink.cleanup()
            self.audio_sink = None
        if metadata.get("end-of-stream"):
            log("client sent end-of-stream, closing audio pipeline")
            self.stop_receiving_audio()
            return
        if not self.audio_sink:
            if not self.audio_loop_check("microphone"):
                # make a fake object,
                # so we don't fire the audio loop check warning repeatedly
                self.audio_sink = FakeSink(codec)
                return
            try:
                def audio_sink_error(*args) -> None:
                    log("audio_sink_error%s", args)
                    log.warn("Warning: stopping audio input because of an error")
                    self.stop_receiving_audio()

                from xpra.audio.wrapper import start_receiving_audio  # pylint: disable=import-outside-toplevel
                ss = start_receiving_audio(codec)
                if not ss:
                    return
                self.audio_sink = ss
                log("audio_data(..) created audio sink: %s", self.audio_sink)
                ss.connect("error", audio_sink_error)
                ss.start()
                log("audio_data(..) audio sink started")
            except Exception:
                log.error("Error: failed to start receiving %r", codec, exc_info=True)
                return
        self.audio_sink.add_data(data, metadata, packet_metadata)

    def get_audio_source_latency(self) -> int:
        encoder_latency = 0
        ss = self.audio_source
        cinfo = ""
        if ss:
            info = typedict(ss.info or {})
            try:
                qdict = info.dictget("queue")
                if qdict:
                    q = typedict(qdict).intget("cur", 0)
                    log("server side queue level: %s", q)
                # get the latency from the source info, if it has it:
                encoder_latency = info.intget("latency", -1)
                if encoder_latency < 0:
                    # fallback to hard-coded values:
                    # pylint: disable=import-outside-toplevel
                    from xpra.audio.gstreamer_util import ENCODER_LATENCY, RECORD_PIPELINE_LATENCY
                    encoder_latency = RECORD_PIPELINE_LATENCY + ENCODER_LATENCY.get(ss.codec, 0)
                    cinfo = f"{ss.codec} "
                # processing overhead
                encoder_latency += 100
            except Exception as e:
                encoder_latency = 0
                log("failed to get encoder latency for %s: %s", ss.codec, e)
        log("get_audio_source_latency() %s: %s", cinfo, encoder_latency)
        return encoder_latency

    def get_info(self) -> dict[str, Any]:
        return {"audio": self.get_audio_info()}

    def get_audio_info(self) -> dict[str, Any]:
        def audio_info(supported, subprocess_wrapper, codecs) -> dict[str, Any]:
            i = {"codecs": codecs}
            if not supported:
                i["state"] = "disabled"
                return i
            if subprocess_wrapper is None:
                i["state"] = "inactive"
                return i
            i.update(subprocess_wrapper.get_info())
            return i

        info = {
            "speaker": audio_info(self.supports_speaker, self.audio_source, self.audio_decoders),
            "microphone": audio_info(self.supports_microphone, self.audio_sink, self.audio_encoders),
        }
        for prop in ("pulseaudio_id", "pulseaudio_server"):
            v = getattr(self, prop)
            if v is not None:
                info[prop] = v
        return info

    def suspend(self) -> None:
        if self.audio_source:
            self.audio_suspended = True
            self.stop_sending_audio()

    def resume(self) -> None:
        suspended = self.audio_suspended
        rsa = self.restart_speaker_args
        if rsa and suspended:
            self.audio_suspended = False
            self.start_sending_audio(*rsa)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import re
import sys

from xpra.auth.sys_auth_base import log
from xpra.auth.sqlauthbase import SQLAuthenticator, DatabaseUtilBase, run_dbutil


def url_path_to_dict(path: str) -> dict[str, str]:
    pattern = (
        r'^'
        r'((?P<schema>.+?)://)?'
        r'((?P<user>.+?)(:(?P<password>.*?))?@)?'
        r'(?P<host>.*?)'
        r'(:(?P<port>\d+?))?'
        r'(?P<path>/.*?)?'
        r'(?P<query>[?].*?)?'
        r'$'
    )
    regex = re.compile(pattern)
    m = regex.match(path)
    return m.groupdict() if m is not None else {}


def db_from_uri(uri: str):
    d = url_path_to_dict(uri)
    log("settings for uri=%s : %s", uri, d)
    import mysql.connector as mysql  # pylint: disable=import-outside-toplevel
    db = mysql.connect(
        host=d.get("host", "localhost"),
        # port = int(d.get("port", 3306)),
        user=d.get("user", ""),
        passwd=d.get("password", ""),
        database=(d.get("path") or "").lstrip("/") or "xpra",
    )
    return db


class Authenticator(SQLAuthenticator):
    CLIENT_USERNAME = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.uri = kwargs.get("uri", "")
        assert self.uri, "missing database uri"

    def db_cursor(self, *sqlargs):
        db = db_from_uri(self.uri)
        cursor = db.cursor()
        cursor.execute(*sqlargs)
        # keep reference to db,
        # so it doesn't get garbage collected just yet:
        cursor.db = db
        log("db_cursor(%s)=%s", sqlargs, cursor)
        return cursor

    def __repr__(self):
        return "mysql"


class MySQLDatabaseUtil(DatabaseUtilBase):

    def __init__(self, uri):
        super().__init__(uri)
        import mysql.connector as mysql
        assert mysql.paramstyle == "pyformat"
        self.param = "%s"

    def exec_database_sql_script(self, cursor_cb, *sqlargs):
        db = db_from_uri(self.uri)
        cursor = db.cursor()
        log("%s.execute%s", cursor, sqlargs)
        cursor.execute(*sqlargs)
        if cursor_cb:
            cursor_cb(cursor)
        db.commit()
        return cursor

    def get_authenticator_class(self) -> type:
        return Authenticator


def main(argv) -> int:
    return run_dbutil(MySQLDatabaseUtil, "databaseURI", argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.gtk.window import add_close_accel

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")

width = 400
height = 200


def make_win():
    window = Gtk.Window(type=Gtk.WindowType.POPUP)
    window.set_title("Main")
    window.set_size_request(width, height)
    window.set_position(Gtk.WindowPosition.CENTER)
    window.connect("delete_event", Gtk.main_quit)
    window.set_events(Gdk.EventMask.KEY_PRESS_MASK | Gdk.EventMask.BUTTON_PRESS_MASK)

    def on_press(*_args):
        Gtk.main_quit()

    window.connect("key_press_event", on_press)
    window.connect("button_press_event", on_press)
    return window


def main():
    with program_context("window-overrideredirect", "Window Override Redirect"):
        w = make_win()
        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("override-redirect test window")
        add_close_accel(w, Gtk.main_quit)

        def show_with_focus():
            force_focus()
            w.show_all()
            w.present()

        GLib.idle_add(show_with_focus)
        Gtk.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from collections.abc import Sequence


PIPE_NAME = "xpra-service"


def get_commonappdata_dir() -> str:
    CSIDL_COMMON_APPDATA = 35
    try:
        from ctypes import (
            create_unicode_buffer, wintypes,
            WinDLL,  # @UnresolvedImport
        )
        buf = create_unicode_buffer(wintypes.MAX_PATH)
        shell32 = WinDLL("shell32", use_last_error=True)
        SHGetFolderPath = shell32.SHGetFolderPathW
        SHGetFolderPath(0, CSIDL_COMMON_APPDATA, None, 0, buf)
        return buf.value
    except Exception:
        return ""


def main(argv: Sequence[str]) -> int:
    if len(argv) > 1 and argv[1] != "start":
        if argv[1] != "stop":
            raise ValueError(f"unsupported subcommand {argv[1]!r}")
        return run_mode("stop", argv[:1] + ["stop", f"named-pipe://{PIPE_NAME}"])

    from multiprocessing import freeze_support
    freeze_support()

    os.environ["XPRA_REDIRECT_OUTPUT"] = "1"
    os.environ["XPRA_LOG_FILENAME"] = "System-Proxy.log"
    # os.environ["XPRA_ALL_DEBUG"] = "1"

    args = argv[:1] + [
        "proxy",
        "--bind-tcp=0.0.0.0:14500,auth=sys,client-username=true,verify-username=true",
        "--tray=no",
        # "-d", "win32,proxy",
        # "--mdns=no",
    ]
    SYSTEM_USER = os.environ.get("SYSTEM_USER", "1")
    if SYSTEM_USER == "1":
        # only SYSTEM can access this named pipe:
        # (so no need for auth)
        os.environ["XPRA_NAMED_PIPE_UNRESTRICTED"] = "0"
        args.append(f"--bind={PIPE_NAME}")
    else:
        # public named pipe (needs auth):
        os.environ["XPRA_NAMED_PIPE_UNRESTRICTED"] = "1"
        args.append(f"--bind={PIPE_NAME},auth=sys")
    commonappdata = get_commonappdata_dir()
    if commonappdata:
        ssl_cert = os.path.join(commonappdata, "Xpra", "ssl-cert.pem")
        if os.path.exists(ssl_cert):
            args.append(f"--ssl-cert={ssl_cert}")
    return run_mode("Xpra-Proxy", args)


def run_mode(name: str, args: Sequence[str]) -> int:
    from xpra.platform import init, set_default_name
    set_default_name(name)
    init()
    from xpra.scripts.main import main as xpra_main
    return int(xpra_main(args[0], args))


if __name__ == "__main__":
    import sys

    r = main(sys.argv)
    sys.exit(r)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import time
from typing import Any
from collections.abc import Sequence

from xpra.os_util import gi_import
from xpra.common import PaintCallbacks
from xpra.client.gui.widget_base import ClientWidgetBase
from xpra.client.gui.window_backing_base import WindowBackingBase, fire_paint_callbacks
from xpra.util.objects import typedict
from xpra.util.str_fn import memoryview_to_bytes
from xpra.util.env import envbool
from xpra.log import Logger

log = Logger("tray")

GLib = gi_import("GLib")

SAVE = envbool("XPRA_SAVE_SYSTRAY", False)


class ClientTray(ClientWidgetBase):
    """
        This acts like a widget, we use the `TrayBacking`
        to capture the tray pixels and forward them
        to the real tray widget class.
    """
    DEFAULT_LOCATION = [0, 0]
    DEFAULT_SIZE = [64, 64]
    DEFAULT_GEOMETRY = DEFAULT_LOCATION + DEFAULT_SIZE

    def __init__(self, client, wid, w, h, metadata, tray_widget, mmap_area):
        log("ClientTray%s", (client, wid, w, h, tray_widget, mmap_area))
        super().__init__(client, 0, wid, True)
        self._size = w, h
        self._metadata = metadata
        self.title = metadata.strget("title")
        self.tray_widget = tray_widget
        self._geometry = None
        self._window_alpha = True
        self.group_leader = None

        self.mmap = mmap_area
        self.new_backing(w, h)
        GLib.idle_add(self.reconfigure)
        # things may have settled by now
        GLib.timeout_add(1000, self.send_configure)

    def reset_size_constraints(self) -> None:
        """ we can't implement this method for trays - it should not be called """

    def resize(self, nw, nh) -> None:
        """ we can't implement this method for trays - it should not be called """

    def set_alpha(self) -> None:
        """
        nothing to do,
        trays aren't really windows and transparency is always supported
        """

    def get_backing_class(self) -> type:
        return TrayBacking

    def is_OR(self) -> bool:
        return True

    def is_tray(self) -> bool:
        return True

    def get_window(self):
        return None

    def get_geometry(self):
        return self._geometry or ClientTray.DEFAULT_GEOMETRY

    def get_tray_geometry(self):
        tw = self.tray_widget
        if not tw:
            return None
        return tw.get_geometry()

    def get_tray_size(self):
        tw = self.tray_widget
        if not tw:
            return None
        return tw.get_size()

    def freeze(self) -> None:
        """
        System trays are small, no point in freezing anything
        """

    def send_configure(self) -> None:
        self.reconfigure(True)

    def reconfigure(self, force_send_configure=False) -> None:
        geometry = None
        tw = self.tray_widget
        if tw:
            geometry = tw.get_geometry()
        log("%s.reconfigure(%s) geometry=%s", self, force_send_configure, geometry)
        if geometry is None:
            if self._geometry or not tw:
                geometry = self.get_geometry()
            else:
                # make one up as best we can - maybe we have the size at least?
                size = tw.get_size()
                log("%s.reconfigure() guessing location using size=%s", self, size)
                geometry = ClientTray.DEFAULT_LOCATION + list(size or ClientTray.DEFAULT_SIZE)
        x, y, w, h = geometry
        if w <= 1 or h <= 1:
            w, h = ClientTray.DEFAULT_SIZE
            geometry = x, y, w, h
        if force_send_configure or self._geometry is None or geometry != self._geometry:
            self._geometry = geometry
            client_properties = {
                "encoding.transparency": True,
                "encodings.rgb_formats": ["RGBA", "RGB", "RGBX"],
            }
            if tw:
                orientation = tw.get_orientation()
                if orientation:
                    client_properties["orientation"] = orientation
            # scale to server coordinates
            sx, sy, sw, sh = self._client.crect(x, y, w, h)
            log("%s.reconfigure(%s) sending configure for geometry=%s : %s",
                self, force_send_configure, geometry, (sx, sy, sw, sh, client_properties))
            self._client.send("configure-window", self.wid, sx, sy, sw, sh, client_properties)
        if self._size != (w, h):
            self.new_backing(w, h)

    def move_resize(self, x: int, y: int, w: int, h: int) -> None:
        log("%s.move_resize(%s, %s, %s, %s)", self, x, y, w, h)
        w = max(1, w)
        h = max(1, h)
        self._geometry = x, y, w, h
        self.reconfigure(True)

    def new_backing(self, w: int, h: int) -> None:
        self._size = w, h
        data = None
        if self._backing:
            data = self._backing.data
        self._backing = TrayBacking(self.wid, w, h, self._has_alpha, data)
        if self.mmap:
            self._backing.enable_mmap(self.mmap)

    def update_metadata(self, metadata) -> None:
        log("%s.update_metadata(%s)", self, metadata)

    def update_icon(self, img) -> None:
        """
        this is the window icon... not the tray icon!
        ignore it as it is never shown anywhere
        """

    def draw_region(self, x: int, y: int, width: int, height: int,
                    coding: str, img_data, rowstride: int, options: typedict,
                    callbacks: PaintCallbacks):
        log("%s.draw_region%s", self,
            (x, y, width, height, coding, "%s bytes" % len(img_data), rowstride, options, callbacks))

        # note: a new backing may be assigned between the time we call draw_region
        # and the time we get the callback (as the draw may use idle_add)
        backing = self._backing

        def after_draw_update_tray(success: bool | int, message: str = "") -> None:
            log("%s.after_draw_update_tray(%s, %s)", self, success, message)
            if not success:
                log.warn(f"Warning: tray paint update failed: {message!r}")
                log.warn(f" for {width}x{height} {coding} update with {options=}")
                return
            tray_data = backing.data
            log("tray backing=%s, data: %s", backing, tray_data is not None)
            if tray_data is None:
                log.warn("Warning: no pixel data in tray backing for window %i", backing.wid)
                return
            GLib.idle_add(self.set_tray_icon, *tray_data)
            GLib.idle_add(self.reconfigure)

        callbacks.append(after_draw_update_tray)
        backing.draw_region(x, y, width, height, coding, img_data, rowstride, options, callbacks)

    def set_tray_icon(self, rgb_format: str, w: int, h: int, rowstride: int, pixels, options: typedict) -> None:
        log("%s.set_tray_icon(%s, %s, %s, %s, %s bytes)", self, rgb_format, w, h, rowstride, len(pixels))
        has_alpha = rgb_format.find("A") >= 0
        tw = self.tray_widget
        if tw:
            # some tray implementations can't deal with memoryviews..
            if isinstance(pixels, (memoryview, bytearray)):
                pixels = memoryview_to_bytes(pixels)
            tw.set_icon_from_data(pixels, has_alpha, w, h, rowstride, options)

    def destroy(self) -> None:
        tw = self.tray_widget
        if tw:
            self.tray_widget = None
            tw.cleanup()

    def __repr__(self):
        return f"ClientTray({self.wid}:{self.title})"


class TrayBacking(WindowBackingBase):
    """
        This backing only stores the rgb pixels,
        so that we can use them with the real widget.
    """

    # keep it simple: only accept 32-bit RGB(X),
    # all tray implementations support alpha
    RGB_MODES: Sequence[str] = ("RGBA", "RGBX")
    HAS_ALPHA = True

    def __init__(self, wid: int, _w: int, _h: int, _has_alpha: bool, data=None):
        self.data = data
        super().__init__(wid, True)
        self._backing = object()  # pretend we have a backing structure

    def get_encoding_properties(self) -> dict[str, Any]:
        # override so we skip all csc caps:
        return {
            "encodings.rgb_formats": self.get_rgb_formats(),
            "encoding.transparency": True,
        }

    def paint_scroll(self, img_data, options, callbacks) -> None:
        raise RuntimeError("scroll should not be used with tray icons")

    def do_paint_rgb(self, context, encoding: str, rgb_format: str, img_data,
                     x: int, y: int, width: int, height: int, render_width: int, render_height: int, rowstride: int,
                     options: typedict, callbacks: PaintCallbacks) -> None:
        if width != render_width or height != render_height:
            fire_paint_callbacks(callbacks, False, "tray paint must not use scaling")
            return
        self.data = (rgb_format, width, height, rowstride, img_data[:], options)
        if SAVE:
            self.save_tray_png()
        fire_paint_callbacks(callbacks, True)

    def save_tray_png(self) -> None:
        log("save_tray_png()")
        rgb_mode, width, height, stride, img_data = self.data[:5]
        mode = "RGB"
        data_mode = "RGB"
        if rgb_mode == "rgb32":
            mode += "A"
            data_mode += "A"
        if not stride:
            stride = width * len(data_mode)
        try:
            from PIL import Image  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            log(f"cannot save tray: {e}")
            return
        img = Image.frombytes(mode, (width, height), img_data, "raw", data_mode, stride, 1)
        filename = f"./tray-{rgb_mode}-{time()}.png"
        img.save(filename, "PNG")
        log.info("tray %s update saved to %s", rgb_mode, filename)



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.str_fn import strtobytes, hexstr
from xpra.net.rfb.protocol import RFBProtocol
from xpra.net.rfb.const import RFBAuth, AUTH_STR, CLIENT_INIT
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("network", "protocol", "rfb")
authlog = Logger("auth", "rfb")


class RFBServerProtocol(RFBProtocol):

    def __init__(self, conn, auth, process_packet_cb, get_rfb_pixelformat, session_name="Xpra", data=b""):
        self._authenticator = auth
        self._get_rfb_pixelformat = get_rfb_pixelformat
        self.session_name = session_name
        super().__init__(conn, process_packet_cb, data=data)

    def handshake_complete(self):
        log.info("RFB connection from %s", self._conn.target)
        # reply with Security Handshake:
        self._packet_parser = self._parse_security_handshake
        if self._authenticator and self._authenticator.requires_challenge():
            security_types = [RFBAuth.VNC]
        else:
            security_types = [RFBAuth.NONE]
        packet = struct.pack(b"B", len(security_types))
        for x in security_types:
            packet += struct.pack(b"B", x)
        self.send(packet)

    def _parse_security_handshake(self, packet):
        authlog("parse_security_handshake(%s)", hexstr(packet))
        try:
            auth = struct.unpack(b"B", packet)[0]
        except struct.error:
            self._internal_error(packet, "cannot parse security handshake response '%s'" % hexstr(packet))
            return 0
        auth_str = AUTH_STR.get(auth, auth)
        if auth == RFBAuth.VNC:
            # send challenge:
            self._packet_parser = self._parse_challenge
            assert self._authenticator
            challenge, digest = self._authenticator.get_challenge(("des", ))
            assert digest == "des", "invalid digest %r, only 'des' is supported" % digest
            self._challenge = challenge[:16]
            authlog("sending RFB challenge value: %s", hexstr(self._challenge))
            self.send(self._challenge)
            return 1
        if self._authenticator and self._authenticator.requires_challenge():
            self.invalid_header(self, packet, "invalid security handshake response, authentication is required")
            return 0
        authlog("parse_security_handshake: auth=%s, sending SecurityResult", auth_str)
        # Security Handshake, send SecurityResult Handshake
        self._packet_parser = self._parse_security_result
        self.send(struct.pack(b"!I", 0))
        return 1

    def _parse_challenge(self, response):
        authlog("parse_challenge(%s)", hexstr(response))
        if len(response) < 16:
            return 0
        assert self._authenticator
        try:
            assert len(response) == 16
            hex_response = hexstr(response)
            # log("padded password=%s", password)
            caps = typedict({
                "challenge_response": hex_response,
            })
            if self._authenticator.authenticate(caps):
                authlog("challenge authentication succeeded")
                self.send(struct.pack(b"!I", 0))
                self._packet_parser = self._parse_security_result
                return 16
            authlog.warn("Warning: authentication challenge response failure")
            authlog.warn(" password does not match")
        except Exception as e:
            authlog("parse_challenge(%s)", hexstr(response), exc_info=True)
            authlog.error("Error: authentication challenge failure:")
            authlog.estr(e)
        GLib.timeout_add(1000, self.send_fail_challenge)
        return len(response)

    def send_fail_challenge(self):
        self.send(struct.pack(b"!I", 1))
        self.close()

    def _parse_security_result(self, packet):
        self.share = packet != b"\0"
        authlog("parse_security_result: sharing=%s, sending ClientInit with session-name=%s",
                self.share, self.session_name)
        # send ServerInit
        self._packet_parser = self._parse_rfb
        try:
            sn = self.session_name.encode("utf8")
        except UnicodeEncodeError:
            sn = strtobytes(self.session_name)
        w, h, bpp, depth, bigendian, truecolor, rmax, gmax, bmax, rshift, bshift, gshift = self._get_rfb_pixelformat()
        packet = struct.pack(CLIENT_INIT,
                             w, h, bpp, depth, bigendian, truecolor,
                             rmax, gmax, bmax, rshift, bshift, gshift,
                             0, 0, 0, len(sn)) + sn
        self.send(packet)
        self._process_packet_cb(self, [b"authenticated"])
        return 1



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from typing import Any
from collections.abc import Callable
import cairo

from xpra.common import noop
from xpra.gtk.util import get_root_size, get_default_root_window
from xpra.os_util import WIN32, gi_import
from xpra.util.env import envint, envbool, IgnoreWarningsContext
from xpra.log import Logger, consume_verbose_argv

Gdk = gi_import("Gdk")
GObject = gi_import("GObject")

log = Logger("gtk", "util")
screenlog = Logger("gtk", "screen")

SHOW_ALL_VISUALS = False
GTK_WORKAREA = envbool("XPRA_GTK_WORKAREA", True)


def get_screen_info(display, screen) -> dict[str, Any]:
    info = {}
    if not WIN32:
        try:
            w = screen.get_root_window()
            if w:
                info["root"] = w.get_geometry()
        except Exception:
            pass
    with IgnoreWarningsContext():
        info["name"] = screen.make_display_name()
    with IgnoreWarningsContext():
        for x in ("width", "height", "width_mm", "height_mm", "resolution", "primary_monitor"):
            fn = getattr(screen, "get_" + x)
            try:
                info[x] = int(fn())
            except Exception:
                pass
    info["monitors"] = display.get_n_monitors()
    m_info = info.setdefault("monitor", {})
    for i in range(display.get_n_monitors()):
        monitor = display.get_monitor(i)
        m_info[i] = get_monitor_info(monitor)
    fo = screen.get_font_options()
    # win32 and osx return nothing here...
    if fo:
        fontoptions = info.setdefault("fontoptions", {})
        fontoptions.update(get_font_info(fo))
    vinfo = info.setdefault("visual", {})

    def visual(name, v):
        i = get_visual_info(v)
        if i:
            vinfo[name] = i

    visual("rgba", screen.get_rgba_visual())
    visual("system_visual", screen.get_system_visual())
    if SHOW_ALL_VISUALS:
        for i, v in enumerate(screen.list_visuals()):
            visual(i, v)

    # Gtk.settings
    def get_setting(key: str, gtype):
        v = GObject.Value()
        v.init(gtype)
        if screen.get_setting(key, v):
            return v.get_value()
        return None

    sinfo = info.setdefault("settings", {})
    for x, gtype in {
        # NET:
        "enable-event-sounds": GObject.TYPE_INT,
        "icon-theme-name": GObject.TYPE_STRING,
        "sound-theme-name": GObject.TYPE_STRING,
        "theme-name": GObject.TYPE_STRING,
        # Xft:
        "xft-antialias": GObject.TYPE_INT,
        "xft-dpi": GObject.TYPE_INT,
        "xft-hinting": GObject.TYPE_INT,
        "xft-hintstyle": GObject.TYPE_STRING,
        "xft-rgba": GObject.TYPE_STRING,
    }.items():
        try:
            v = get_setting("gtk-" + x, gtype)
        except Exception:
            screenlog("failed to query screen '%s'", x, exc_info=True)
            continue
        if v is None:
            v = ""
        if x.startswith("xft-"):
            x = x[4:]
        sinfo[x] = v
    return info


FONT_CONV: dict[str, dict[Any, Any]] = {
    "antialias": {
        cairo.ANTIALIAS_DEFAULT: "default",
        cairo.ANTIALIAS_NONE: "none",
        cairo.ANTIALIAS_GRAY: "gray",
        cairo.ANTIALIAS_SUBPIXEL: "subpixel",
    },
    "hint_metrics": {
        cairo.HINT_METRICS_DEFAULT: "default",
        cairo.HINT_METRICS_OFF: "off",
        cairo.HINT_METRICS_ON: "on",
    },
    "hint_style": {
        cairo.HINT_STYLE_DEFAULT: "default",
        cairo.HINT_STYLE_NONE: "none",
        cairo.HINT_STYLE_SLIGHT: "slight",
        cairo.HINT_STYLE_MEDIUM: "medium",
        cairo.HINT_STYLE_FULL: "full",
    },
    "subpixel_order": {
        cairo.SUBPIXEL_ORDER_DEFAULT: "default",
        cairo.SUBPIXEL_ORDER_RGB: "RGB",
        cairo.SUBPIXEL_ORDER_BGR: "BGR",
        cairo.SUBPIXEL_ORDER_VRGB: "VRGB",
        cairo.SUBPIXEL_ORDER_VBGR: "VBGR",
    }
}


def get_font_info(font_options: cairo.FontOptions) -> dict[str, Any]:
    # pylint: disable=no-member
    font_info: dict[str, Any] = {}
    for x, vdict in FONT_CONV.items():
        fn = getattr(font_options, "get_" + x)
        val = fn()
        font_info[x] = vdict.get(val, val)
    return font_info


VISUAL_NAMES: dict[Gdk.VisualType, str] = {
    Gdk.VisualType.STATIC_GRAY: "STATIC_GRAY",
    Gdk.VisualType.GRAYSCALE: "GRAYSCALE",
    Gdk.VisualType.STATIC_COLOR: "STATIC_COLOR",
    Gdk.VisualType.PSEUDO_COLOR: "PSEUDO_COLOR",
    Gdk.VisualType.TRUE_COLOR: "TRUE_COLOR",
    Gdk.VisualType.DIRECT_COLOR: "DIRECT_COLOR",
}
BYTE_ORDER_NAMES: dict[Gdk.ByteOrder, str] = {
    Gdk.ByteOrder.LSB_FIRST: "LSB",
    Gdk.ByteOrder.MSB_FIRST: "MSB",
}
SUBPIXEL_LAYOUT: dict[Gdk.SubpixelLayout, str] = {
    Gdk.SubpixelLayout.UNKNOWN: "unknown",
    Gdk.SubpixelLayout.NONE: "none",
    Gdk.SubpixelLayout.HORIZONTAL_RGB: "horizontal-rgb",
    Gdk.SubpixelLayout.HORIZONTAL_BGR: "horizontal-bgr",
    Gdk.SubpixelLayout.VERTICAL_RGB: "vertical-rgb",
    Gdk.SubpixelLayout.VERTICAL_BGR: "vertical-bgr",
}
VINFO_CONV: dict[str, dict[Any, str]] = {
    "bits_per_rgb": {},
    "byte_order": BYTE_ORDER_NAMES,
    "colormap_size": {},
    "depth": {},
    "red_pixel_details": {},
    "green_pixel_details": {},
    "blue_pixel_details": {},
    "visual_type": VISUAL_NAMES,
}


def get_visual_info(v) -> dict[str, Any]:
    if not v:
        return {}
    vinfo: dict[str, Any] = {}
    for x, vdict in VINFO_CONV.items():
        try:
            fn = getattr(v, "get_" + x)
        except AttributeError:
            pass
        else:
            with IgnoreWarningsContext():
                val = fn()
            if val is not None:
                vinfo[x] = vdict.get(val, val)
    return vinfo


def get_rectangle_info(rect: Gdk.Rectangle) -> dict[str, int]:
    info: dict[str, int] = {}
    for x in ("x", "y", "width", "height"):
        info[x] = getattr(rect, x)
    return info


def get_monitor_info(monitor: Gdk.Monitor) -> dict[str, Any]:
    geom = monitor.get_geometry()
    info: dict[str, Any] = get_rectangle_info(geom)
    for x in ("manufacturer", "model"):
        info[x] = getattr(monitor, f"get_{x}", noop)() or ""
    for x in ("scale_factor", "width_mm", "height_mm", "refresh_rate"):
        if hasattr(monitor, f"get_{x}"):
            fn: Callable = getattr(monitor, f"get_{x}")
            info[x] = int(fn())
    workarea = monitor.get_workarea()
    info["workarea"] = get_rectangle_info(workarea)
    subpixel = monitor.get_subpixel_layout()
    info["subpixel-layout"] = SUBPIXEL_LAYOUT.get(subpixel, "unknown")
    return info


def get_monitors_info(xscale: float = 1.0, yscale: float = 1.0) -> dict[int, Any]:
    display = Gdk.Display.get_default()
    info: dict[int, Any] = {}
    n = display.get_n_monitors()
    for i in range(n):
        minfo = info.setdefault(i, {})
        monitor = display.get_monitor(i)
        minfo["primary"] = monitor.is_primary()
        for attr in (
                "geometry", "refresh-rate", "scale-factor",
                "width-mm", "height-mm",
                "manufacturer", "model",
                "subpixel-layout", "workarea",
        ):
            getter = getattr(monitor, "get_%s" % attr.replace("-", "_"), None)
            if getter:
                value = getter()
                if value is None:
                    continue
                if isinstance(value, Gdk.Rectangle):
                    value = (
                        round(xscale * value.x),
                        round(yscale * value.y),
                        round(xscale * value.width),
                        round(yscale * value.height),
                    )
                elif attr == "width-mm":
                    value = round(xscale * value)
                elif attr == "height-mm":
                    value = round(yscale * value)
                elif attr == "subpixel-layout":
                    value = {
                        Gdk.SubpixelLayout.UNKNOWN: "unknown",
                        Gdk.SubpixelLayout.NONE: "none",
                        Gdk.SubpixelLayout.HORIZONTAL_RGB: "horizontal-rgb",
                        Gdk.SubpixelLayout.HORIZONTAL_BGR: "horizontal-bgr",
                        Gdk.SubpixelLayout.VERTICAL_RGB: "vertical-rgb",
                        Gdk.SubpixelLayout.VERTICAL_BGR: "vertical-bgr",
                    }.get(value, "unknown")
                if isinstance(value, str):
                    value = value.strip()
                minfo[attr] = value
    return info


def get_display_info(xscale=1, yscale=1) -> dict[str, Any]:
    display = Gdk.Display.get_default()

    def xy(v) -> tuple[int, int]:
        return round(xscale * v[0]), round(yscale * v[1])

    def avg(v) -> int:
        return round((xscale * v + yscale * v) / 2)

    root_size = get_root_size()
    with IgnoreWarningsContext():
        info: dict[str, Any] = {
            "root-size": xy(root_size),
            "name": display.get_name(),
            "pointer": xy(display.get_pointer()[-3:-1]),
            "devices": len(display.list_devices()),
            "default_cursor_size": avg(display.get_default_cursor_size()),
            "maximal_cursor_size": xy(display.get_maximal_cursor_size()),
            "pointer_is_grabbed": display.pointer_is_grabbed(),
        }
    if not WIN32:
        rw = get_default_root_window()
        if rw:
            info["root"] = rw.get_geometry()
    sinfo = info.setdefault("supports", {})
    for x in ("composite", "cursor_alpha", "cursor_color", "selection_notification", "clipboard_persistence", "shapes"):
        f = "supports_" + x
        if hasattr(display, f):
            fn = getattr(display, f)
            with IgnoreWarningsContext():
                sinfo[x] = fn()
    info["screens"] = get_screens_info()
    info["monitors"] = get_monitors_info(xscale, yscale)
    with IgnoreWarningsContext():
        dm = display.get_device_manager()
        for dt, name in {
            Gdk.DeviceType.MASTER: "master",
            Gdk.DeviceType.SLAVE: "slave",
            Gdk.DeviceType.FLOATING: "floating",
        }.items():
            dinfo = info.setdefault("device", {})
            dtinfo = dinfo.setdefault(name, {})
            devices = dm.list_devices(dt)
            for i, d in enumerate(devices):
                dtinfo[i] = d.get_name()
    return info


def get_screens_info() -> dict[int, dict]:
    display = Gdk.Display.get_default()
    info: dict[int, dict] = {}
    with IgnoreWarningsContext():
        assert display.get_n_screens() == 1, "GTK3: The number of screens is always 1"
        screen = display.get_screen(0)
    info[0] = get_screen_info(display, screen)
    return info


def get_screen_sizes(xscale: float = 1, yscale: float = 1):
    from xpra.platform.gui import get_workarea, get_workareas

    def xs(v) -> int:
        return round(v / xscale)

    def ys(v) -> int:
        return round(v / yscale)

    def swork(*workarea) -> tuple[int, int, int, int]:
        return xs(workarea[0]), ys(workarea[1]), xs(workarea[2]), ys(workarea[3])

    display = Gdk.Display.get_default()
    if not display:
        return []
    MIN_DPI = envint("XPRA_MIN_DPI", 10)
    MAX_DPI = envint("XPRA_MIN_DPI", 500)

    def dpi(size_pixels: int, size_mm: int) -> int:
        if size_mm == 0:
            return 0
        return round(size_pixels * 254 / size_mm / 10)

    # GTK 3.22 onwards always returns just a single screen,
    # potentially with multiple monitors
    n_monitors = display.get_n_monitors()
    workareas = get_workareas()
    if workareas and len(workareas) != n_monitors:
        screenlog(" workareas: %s", workareas)
        screenlog(" number of monitors does not match number of workareas!")
        workareas = []
    monitors = []
    for j in range(n_monitors):
        monitor = display.get_monitor(j)
        geom = monitor.get_geometry()
        manufacturer, model = monitor.get_manufacturer(), monitor.get_model()
        if manufacturer in ("unknown", None):
            manufacturer = ""
        if model in ("unknown", None):
            model = ""
        if manufacturer and model:
            plug_name = "%s %s" % (manufacturer, model)
        elif manufacturer:
            plug_name = manufacturer
        elif model:
            plug_name = model
        else:
            plug_name = "%i" % j
        wmm, hmm = monitor.get_width_mm(), monitor.get_height_mm()
        monitor_info = [plug_name, xs(geom.x), ys(geom.y), xs(geom.width), ys(geom.height), wmm, hmm]
        screenlog(" monitor %s: %s, model=%s, manufacturer=%s",
                  j, type(monitor).__name__, monitor.get_model(), monitor.get_manufacturer())

        def vmwx(v) -> bool:
            return v < geom.x or v > geom.x + geom.width

        def vmwy(v) -> bool:
            return v < geom.y or v > geom.y + geom.height

        def valid_workarea(work_x, work_y, work_width, work_height) -> list[int]:
            if vmwx(work_x) or vmwx(work_x + work_width) or vmwy(work_y) or vmwy(work_y + work_height):
                screenlog("discarding invalid workarea: %s", (work_x, work_y, work_width, work_height))
                return []
            return list(swork(work_x, work_y, work_width, work_height))

        if GTK_WORKAREA and hasattr(monitor, "get_workarea"):
            rect = monitor.get_workarea()
            monitor_info += valid_workarea(rect.x, rect.y, rect.width, rect.height)
        elif workareas:
            monitor_info += valid_workarea(*workareas[j])
        monitors.append(tuple(monitor_info))
    screen = display.get_default_screen()
    with IgnoreWarningsContext():
        sw, sh = screen.get_width(), screen.get_height()
    work_x, work_y, work_width, work_height = swork(0, 0, sw, sh)
    workarea = get_workarea()  # pylint: disable=assignment-from-none
    screenlog(" workarea=%s", workarea)
    if workarea:
        work_x, work_y, work_width, work_height = swork(*workarea)  # pylint: disable=not-an-iterable

        def vwx(v) -> bool:
            return v < 0 or v > sw

        def vwy(v) -> bool:
            return v < 0 or v > sh

        if vwx(work_x) or vwx(work_x + work_width) or vwy(work_y) or vwy(work_y + work_height):
            screenlog(" discarding invalid workarea values: %s", workarea)
            work_x, work_y, work_width, work_height = swork(0, 0, sw, sh)
    with IgnoreWarningsContext():
        wmm = screen.get_width_mm()
        hmm = screen.get_height_mm()
    xdpi = dpi(sw, wmm)
    ydpi = dpi(sh, hmm)
    if xdpi < MIN_DPI or xdpi > MAX_DPI or ydpi < MIN_DPI or ydpi > MAX_DPI:
        screenlog(f"ignoring invalid DPI {xdpi},{ydpi} from screen size {wmm}x{hmm}mm")
        if os.environ.get("WAYLAND_DISPLAY"):
            screenlog(" (wayland display?)")
        if n_monitors > 0:
            wmm = 0
            for mi in range(n_monitors):
                monitor = display.get_monitor(mi)
                screenlog(" monitor %i: %s, model=%s, manufacturer=%s",
                          mi, monitor, monitor.get_model(), monitor.get_manufacturer())
                wmm += monitor.get_width_mm()
                hmm += monitor.get_height_mm()
            wmm /= n_monitors
            hmm /= n_monitors
            xdpi = dpi(sw, wmm)
            ydpi = dpi(sh, hmm)
        if xdpi < MIN_DPI or xdpi > MAX_DPI or ydpi < MIN_DPI or ydpi > MAX_DPI:
            # still invalid, generate one from DPI=96
            wmm = round(sw * 25.4 / 96)
            hmm = round(sh * 25.4 / 96)
        screenlog(" using %ix%i mm", wmm, hmm)
    with IgnoreWarningsContext():
        screen0 = (screen.make_display_name(), xs(sw), ys(sh),
                   wmm, hmm,
                   monitors,
                   work_x, work_y, work_width, work_height)
    screenlog(" screen: %s", screen0)
    return [screen0]


def main() -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.util.str_fn import pver
    from xpra.util.str_fn import print_nested_dict

    def print_version_dict(d: dict, vformat=pver) -> None:
        for k in sorted(d.keys()):
            v = d[k]
            print("* %-48s : %r" % (str(k).replace(".version", "").ljust(12), vformat(v)))

    from xpra.platform import program_context
    with program_context("GTK-Version-Info", "GTK Version Info"):
        from xpra.platform.gui import init as gui_init, ready
        gui_init()
        ready()
        if consume_verbose_argv(sys.argv, "gtk"):
            global SHOW_ALL_VISUALS
            SHOW_ALL_VISUALS = True
        print("GTK Version:")
        from xpra.gtk import versions
        print_version_dict(versions.get_gtk_version_info())
        print("Display:")
        print_nested_dict(get_display_info(), vformat=str)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from queue import SimpleQueue, Empty
from typing import Any
from collections.abc import Callable

from xpra.util.parsing import parse_simple_dict
from xpra.util.objects import typedict
from xpra.util.env import envint
from xpra.os_util import OSX, gi_import
from xpra.gstreamer.common import import_gst, GST_FLOW_OK
from xpra.gstreamer.pipeline import Pipeline
from xpra.log import Logger

Gst = import_gst()
log = Logger("encoder", "gstreamer")

FRAME_JUNK_TIMEOUT = envint("XPRA_FRAME_JUNK_TIMEOUT", 100)
FRAME_QUEUE_TIMEOUT = envint("XPRA_GSTREAMER_FRAME_QUEUE_TIMEOUT", 1000)
FRAME_QUEUE_INITIAL_TIMEOUT = envint("XPRA_GSTREAMER_FRAME_QUEUE_INITIAL_TIMEOUT", 3000)


def get_default_encoder_options() -> dict[str, dict[str, Any]]:
    options: dict[str, dict[str, Any]] = {
        "vaapih264enc": {
            "max-bframes": 0,  # int(options.boolget("b-frames", False))
            # "tune"          : 3,    #low-power
            # "rate-control" : 8, #qvbr
            "compliance-mode": 0,  # restrict-buf-alloc (1) – Restrict the allocation size of coded-buffer
            # "keyframe-period"   : 9999,
            # "prediction-type" : 1, #hierarchical-p (1) – Hierarchical P frame encode
            # "quality-factor" : 10,
            # "quality-level" : 50,
            # "bitrate"   : 2000,
            # "prediction-type" : 1,    #Hierarchical P frame encode
            # "keyframe-period" : 4294967295,
            "aud": True,
        },
        "vaapih265enc": {
            "max-bframes": 0,  # int(options.boolget("b-frames", False))
            # "tune"          : 3,    #low-power
            # "rate-control" : 8, #qvbr
        },
        "amfh264enc": {
            "usage": "ultra-low-latency",
        },
        "amfh265enc": {
            "usage": "ultra-low-latency",
        },
        "x264enc": {
            "speed-preset": "ultrafast",
            "tune": "zerolatency",
            "byte-stream": True,
            "threads": 1,
            "key-int-max": 15,
            "intra-refresh": True,
        },
        "vp8enc": {
            "deadline": 1,
            "error-resilient": 0,
        },
        "vp9enc": {
            "deadline": 1,
            "error-resilient": 0,
            "lag-in-frames": 0,
            "cpu-used": 16,
        },
        "nvh264enc": {
            "zerolatency": True,
            "rc-mode": 3,  # vbr
            "preset": 5,  # low latency, high performance
            "bframes": 0,
            "aud": True,
        },
        "nvh265enc": {
            "zerolatency": True,
            "rc-mode": 3,  # vbr
            "preset": 5,  # low latency, high performance
            # should be in GStreamer 1.18, but somehow missing?
            # "bframes"       : 0,
            "aud": True,
        },
        "nvd3d11h264enc": {
            "bframes": 0,
            "aud": True,
            "preset": 5,  # low latency, high performance
            "zero-reorder-delay": True,
        },
        "nvd3d11h265enc": {
            "bframes": 0,
            "aud": True,
            "preset": 5,  # low latency, high performance
            "zero-reorder-delay": True,
        },
        "svtav1enc": {
            # "speed"         : 12,
            # "gop-size"      : 251,
            "intra-refresh": 1,  # open gop
            # "lookahead"     : 0,
            # "rc"            : 1,    #vbr
        },
        "svtvp9enc": {
        },
        # "svthevcenc" : {
        #    "b-pyramid"         : 0,
        #    "baselayer-mode"    : 1,
        #    "enable-open-gop"   : True,
        #    "key-int-max"       : 255,
        #    "lookahead"         : 0,
        #    "pred-struct"       : 0,
        #    "rc"                : 1, #vbr
        #    "speed"             : 9,
        #    "tune"              : 0,
        # }
    }
    if not OSX:
        options["av1enc"] = {
            "cpu-used": 5,
            "end-usage": 2,  # cq
        }
    # now apply environment overrides:
    for element in options.keys():
        enc_options_str = os.environ.get(f"XPRA_{element.upper()}_OPTIONS", "")
        if enc_options_str:
            encoder_options = parse_simple_dict(enc_options_str)
            log(f"user overridden options for {element}: {encoder_options}")
            options[element] = encoder_options
    return options


def get_default_decoder_options() -> dict[str, dict[str, Any]]:
    options: dict[str, dict[str, Any]] = {
        "av1dec": {
            "stream-format": "obu-stream",
            "alignment": "tu",
        },
    }
    return options


def get_version() -> tuple[int, int]:
    return 5, 0


def get_type() -> str:
    return "gstreamer"


def get_info() -> dict[str, Any]:
    return {"version": get_version()}


IDENTICAL_PIXEL_FORMATS = (
    "NV12",
    "RGBA", "BGRA", "ARGB", "ABGR",
    "RGB", "BGR",
    "RGB15", "RGB16", "BGR15",
    "r210",
    "BGRP", "RGBP",
)
XPRA_TO_GSTREAMER = {
    "YUV420P": "I420",
    "YUV444P": "Y444",
    "BGRX": "BGRx",
    "XRGB": "xRGB",
    "XBGR": "xBGR",
    "YUV400": "GRAY8",
    # "RGB8P"
}
GSTREAMER_TO_XPRA = dict((v, k) for k, v in XPRA_TO_GSTREAMER.items())


def get_gst_rgb_format(rgb_format: str) -> str:
    if rgb_format in IDENTICAL_PIXEL_FORMATS:
        return rgb_format
    # translate to gstreamer name:
    return XPRA_TO_GSTREAMER.get(rgb_format, "")


def get_xpra_rgb_format(rgb_format: str) -> str:
    if rgb_format in IDENTICAL_PIXEL_FORMATS:
        return rgb_format
    # translate to gstreamer name:
    return GSTREAMER_TO_XPRA.get(rgb_format, "")


def gst_to_native(value: Any) -> tuple | str:
    if isinstance(value, Gst.ValueList):
        return tuple(value.get_value(value, i) for i in range(value.get_size(value)))
    if isinstance(value, Gst.IntRange):
        return value.range.start, value.range.stop
    if isinstance(value, Gst.FractionRange):

        def preferint(v):
            if int(v) == v:
                return int(v)
            return v

        def numdenom(fraction):
            return preferint(fraction.num), preferint(fraction.denom)
        return numdenom(value.start), numdenom(value.stop)
    # print(f"value={value} ({type(value)} - {dir(value)}")
    return Gst.value_serialize(value)


_overrides_verified = False


def verify_gst_overrides() -> bool:
    hasp3g = has_python3_gstreamer()
    global _overrides_verified
    if not hasp3g and not _overrides_verified:
        log.warn("Warning: `python3-gstreamer` is not installed")
        log.warn(" this will prevent the python bindings from working properly")
    _overrides_verified = True
    return hasp3g


def has_python3_gstreamer() -> bool:
    from importlib.util import find_spec
    return bool(find_spec("gi.overrides.Gst"))


def get_encoder_info(element="vp8enc") -> dict:
    """
    Get the element's input information,
    we only retrieve the "SINK" pad
    if it accepts `video/x-raw`.
    Convert Gst types into native types,
    and convert pixel formats to the names used in xpra.
    """
    factory = Gst.ElementFactory.find(element)
    if not factory:
        return {}
    if factory.get_num_pad_templates() == 0:
        return {}
    verify_gst_overrides()
    pads = factory.get_static_pad_templates()
    info = {}
    GLib = gi_import("GLib")
    log(f"get_encoder_info({element}) pads={pads}")
    for pad in pads:
        if pad.direction != Gst.PadDirection.SINK:
            log(f"ignoring {pad.direction}")
            continue
        padtemplate = pad.get()
        caps = padtemplate.get_caps()
        if not caps:
            log(f"pad template {padtemplate} has no caps!")
            continue
        if caps.is_any():
            continue
        if caps.is_empty():
            continue
        for i in range(caps.get_size()):
            structure = caps.get_structure(i)
            if structure.get_name() != "video/x-raw":
                continue

            def add_cap(field, value):
                fname = GLib.quark_to_string(field)
                native_value = gst_to_native(value)
                log(f"{fname}={native_value} ({type(native_value)})")
                if fname == "format":
                    if isinstance(native_value, str):
                        # convert it to a tuple:
                        native_value = native_value,
                    if isinstance(native_value, tuple):
                        xpra_formats = tuple(get_xpra_rgb_format(x) for x in native_value)
                        native_value = tuple(fmt for fmt in xpra_formats if fmt)
                info[fname] = native_value
                return True
            structure.foreach(add_cap)
    return info


def get_video_encoder_caps(encoder: str = "x264enc") -> dict[str, Any]:
    if encoder == "jpeg":
        return {}
    if encoder == "av1enc":
        return {
            "alignment": "tu",
            "stream-format": "obu-stream",
        }
    return {
        "alignment": "au",
        "stream-format": "byte-stream",
    }


def get_video_encoder_options(encoder: str = "x264", profile: str = "", options: typedict | None = None):
    eopts = get_default_encoder_options().get(encoder, {})
    eopts["name"] = "encoder"
    if encoder == "x264enc" and options:
        from xpra.codecs.constants import get_x264_quality, get_x264_preset
        q = get_x264_quality(options.intget("quality", 50), profile)
        s = options.intget("speed", 50)
        eopts |= {
            "pass": "qual",
            "quantizer": q,
            "speed-preset": get_x264_preset(s),
        }
    # should check for "bframes" flag in options?
    return eopts


def get_gst_encoding(encoding: str) -> str:
    if encoding in ("jpeg", "png"):
        return f"image/{encoding}"
    video = {"hevc": "h265"}.get(encoding, encoding)
    return f"video/x-{video}"


class VideoPipeline(Pipeline):
    __generic_signals__: dict[str, tuple] = Pipeline.__generic_signals__.copy()
    """
    Dispatch video encoding or decoding to a gstreamer pipeline
    """

    def __init__(self):
        super().__init__()
        self.encoding = ""
        self.width = 0
        self.height = 0
        self.colorspace = ""
        self.frames = 0
        self.frame_queue: SimpleQueue[Any] = SimpleQueue()
        self.pipeline_str = ""
        self.src: Gst.Element | None = None
        self.sink: Gst.Element | None = None

    def init_context(self, encoding: str, width: int, height: int, colorspace: str, options: typedict) -> None:
        self.encoding = encoding
        self.width = width
        self.height = height
        self.colorspace = colorspace
        self.create_pipeline(options)
        self.src: Gst.Element = self.pipeline.get_by_name("src")
        self.src.set_property("format", Gst.Format.TIME)
        self.sink: Gst.Element = self.pipeline.get_by_name("sink")

        def sh(sig: str, handler: Callable):
            self.element_connect(self.sink, sig, handler)

        sh("new-sample", self.on_new_sample)
        sh("new-preroll", self.on_new_preroll)
        self.start()

    def create_pipeline(self, options):
        raise NotImplementedError()

    def on_message(self, bus, message) -> int:
        if message.type == Gst.MessageType.NEED_CONTEXT and self.pipeline_str.find("vaapi") >= 0:
            log("vaapi is requesting a context")
            return GST_FLOW_OK
        return super().on_message(bus, message)

    def on_new_preroll(self, _appsink) -> int:
        log("new-preroll")
        return GST_FLOW_OK

    def process_buffer(self, buf, options: typedict):
        r = self.src.emit("push-buffer", buf)
        if r != GST_FLOW_OK:
            log.error("Error: unable to push image buffer")
            return None

        junk = options.boolget("junk")
        if junk:
            timeout = FRAME_JUNK_TIMEOUT
        elif self.frames == 0:
            timeout = FRAME_QUEUE_INITIAL_TIMEOUT
        else:
            timeout = FRAME_QUEUE_TIMEOUT
        try:
            return self.frame_queue.get(timeout=timeout/1000)
        except Empty:
            log_fn = log.debug if junk else log.error
            log_fn(f"Error: frame queue timeout after {timeout}ms")
            try:
                btype = type(buf).__qualname__.lower()
                log_fn(f" on {btype!r} of size {buf.get_size()}")
                log_fn(f" of {repr(self)}")
            except AttributeError:
                pass
            for k, v in self.get_info().items():
                log_fn(f" {k:<16}: {v}")
            return None

    def get_info(self) -> dict[str, Any]:
        info: dict[str, Any] = get_info()
        if not self.colorspace:
            return info
        info |= {
            "frames": self.frames,
            "width": self.width,
            "height": self.height,
            "encoding": self.encoding,
            "colorspace": self.colorspace,
            "version": get_version(),
        }
        return info

    def __repr__(self):
        if not self.colorspace:
            return "gstreamer(uninitialized)"
        return f"gstreamer({self.colorspace} - {self.width}x{self.height})"

    def is_ready(self) -> bool:
        return bool(self.colorspace)

    def is_closed(self) -> bool:
        return not bool(self.colorspace)

    def get_encoding(self) -> str:
        return self.encoding

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_type(self) -> str:
        return "gstreamer"

    def clean(self) -> None:
        super().cleanup()
        self.width = 0
        self.height = 0
        self.colorspace = ""
        self.encoding = ""
        self.frames = 0

    def do_emit_info(self) -> None:
        self.emit_info_timer = 0



#!/usr/bin/env python3
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.gtk.window import add_close_accel
from xpra.gtk.pixbuf import get_icon_pixbuf

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")


def make_window():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_title("Window States")
    window.set_size_request(320, 500)
    window.set_position(Gtk.WindowPosition.CENTER)
    window.connect("delete_event", Gtk.main_quit)
    icon = get_icon_pixbuf("ticked.png")
    if icon:
        window.set_icon(icon)
    vbox = Gtk.VBox(homogeneous=False, spacing=0)

    def add_buttons(t1, cb1, t2, cb2):
        hbox = Gtk.HBox(homogeneous=True, spacing=10)
        b1 = Gtk.Button(label=t1)

        def vcb1(*_args):
            cb1()

        b1.connect('clicked', vcb1)
        hbox.pack_start(b1, expand=True, fill=False, padding=5)
        b2 = Gtk.Button(label=t2)

        def vcb2(*_args):
            cb2()

        b2.connect('clicked', vcb2)
        hbox.pack_start(b2, expand=True, fill=False, padding=5)
        vbox.pack_start(hbox, expand=False, fill=False, padding=2)

    add_buttons("maximize", window.maximize, "unmaximize", window.unmaximize)
    # fullscreen-monitors:
    hbox = Gtk.HBox()
    fsm_entry = Gtk.Entry()
    fsm_entry.set_text("0,0,0,0")
    hbox.add(fsm_entry)

    def set_fsm(*_args):
        v = fsm_entry.get_text()
        strs = v.split(",")
        assert len(strs) == 4, "the list of monitors must have 4 items!"
        monitors = [int(x) for x in strs]
        from xpra.platform.gui import set_fullscreen_monitors
        set_fullscreen_monitors(window.get_window(), monitors)

    set_fsm_btn = Gtk.Button(label="Set Fullscreen Monitors")
    set_fsm_btn.connect("clicked", set_fsm)
    hbox.add(set_fsm_btn)
    vbox.pack_start(hbox, expand=False, fill=False, padding=2)
    add_buttons("fullscreen", window.fullscreen, "unfullscreen", window.unfullscreen)

    def decorate():
        window.set_decorated(True)

    def undecorate():
        window.set_decorated(False)

    add_buttons("decorate", decorate, "undecorate", undecorate)
    add_buttons("iconify", window.iconify, "deiconify", window.deiconify)

    def above():
        window.set_keep_above(True)

    def notabove():
        window.set_keep_above(False)

    add_buttons("keep above", above, "not above", notabove)

    def below():
        window.set_keep_below(True)

    def notbelow():
        window.set_keep_below(False)

    add_buttons("keep below", below, "not below", notbelow)
    add_buttons("stick", window.stick, "unstick", window.unstick)

    def skip_pager():
        window.set_skip_pager_hint(True)

    def notskip_pager():
        window.set_skip_pager_hint(False)

    add_buttons("skip pager", skip_pager, "not skip pager", notskip_pager)

    def skip_taskbar():
        window.set_skip_taskbar_hint(True)

    def notskip_taskbar():
        window.set_skip_taskbar_hint(False)

    add_buttons("skip taskbar", skip_taskbar, "not skip taskbar", notskip_taskbar)

    def shade():
        from xpra.platform.gui import set_shaded
        set_shaded(window.get_window(), True)

    def unshade():
        from xpra.platform.gui import set_shaded
        set_shaded(window.get_window(), False)

    add_buttons("shade", shade, "unshade", unshade)

    def modal():
        window.set_modal(True)

    def notmodal():
        window.set_modal(False)

    add_buttons("modal", modal, "not modal", notmodal)

    def window_state(_widget, event):
        STATES = {
            Gdk.WindowState.WITHDRAWN: "withdrawn",
            Gdk.WindowState.ICONIFIED: "iconified",
            Gdk.WindowState.MAXIMIZED: "maximized",
            Gdk.WindowState.STICKY: "sticky",
            Gdk.WindowState.FULLSCREEN: "fullscreen",
            Gdk.WindowState.ABOVE: "above",
            Gdk.WindowState.BELOW: "below",
        }
        # print("window_state(%s, %s)" % (widget, event))
        print("flags: %s" % [STATES[x] for x in STATES.keys() if x & event.new_window_state])

    window.connect("window-state-event", window_state)
    window.add(vbox)
    return window


def main():
    with program_context("window-states", "Window States"):
        w = make_window()
        add_close_accel(w, Gtk.main_quit)
        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("window states test")

        def show_with_focus():
            force_focus()
            w.show_all()
            w.present()

        GLib.idle_add(show_with_focus)
        Gtk.main()
        return 0


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: boundscheck=False, wraparound=False

import struct
from typing import Tuple, Dict

from xpra.util.env import envbool
from xpra.util.str_fn import repr_ellipsized, csv
from xpra.log import Logger
log = Logger("encoding", "scroll")

from xpra.buffers.membuf cimport memalign, buffer_context  # pylint: disable=syntax-error
from xpra.buffers.xxh cimport xxh3
from xpra.util.rectangle import rectangle


cdef int DEBUG = envbool("XPRA_SCROLL_DEBUG", False)


from libc.stdint cimport uint8_t, int16_t, uint16_t, uint32_t, uint64_t, uintptr_t
from libc.stdlib cimport free, malloc
from libc.string cimport memset


MIN_LINE_COUNT = 2

def h(v) -> str:
    return hex(v)[2:].rstrip("L")

cdef inline uint64_t hashtoint64(s):
    return <uint64_t> struct.unpack(b"@L", s)[0]

cdef da(uint64_t *a, uint16_t l):
    return repr_ellipsized(csv([h(a[i]) for i in range(l)]))

cdef dd(uint16_t *d, uint16_t l):
    return csv([h(d[i]) for i in range(l)])


cdef class ScrollData:

    cdef object __weakref__
    #for each distance, keep track of the hit count:
    cdef uint16_t *distances
    cdef uint64_t *a1        #checksums of reference picture
    cdef uint64_t *a2        #checksums of latest picture
    cdef uint8_t matched
    cdef int16_t x
    cdef int16_t y
    cdef uint16_t width
    cdef uint16_t height

    def __cinit__(self, int16_t x=0, int16_t y=0, uint16_t width=0, uint16_t height=0):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __repr__(self):
        return "ScrollDistances(%ix%i)" % (self.width, self.height)

    #only used by the unit tests:
    def test_update(self, arr) -> None:
        if self.a1:
            free(self.a1)
            self.a1 = NULL
        if self.a2:
            self.a1 = self.a2
            self.a2 = NULL
        cdef uint16_t l = len(arr)
        cdef size_t asize = l*(sizeof(uint64_t))
        self.a2 = <uint64_t*> memalign(asize)
        assert self.a2!=NULL, "checksum memory allocation failed"
        for i,v in enumerate(arr):
            self.a2[i] = <uint64_t> abs(v)

    def update(self, pixels, int16_t x, int16_t y, uint16_t width, uint16_t height,
               uint32_t rowstride, uint8_t bpp=4) -> None:
        """
            Add a new image to compare with,
            checksum its rows into a2,
            and push existing values (if we had any) into a1.
        """
        if DEBUG:
            log("%s.update%s a1=%#x, a2=%#x, distances=%#x, current size: %ix%i", self, (repr_ellipsized(pixels), x, y, width, height, rowstride, bpp), <uintptr_t> self.a1, <uintptr_t> self.a2, <uintptr_t> self.distances, self.width, self.height)
        assert width>0 and height>0, "invalid dimensions: %ix%i" % (width, height)
        #scroll area can move within the window:
        self.x = x
        self.y = y
        #but cannot change size (checksums would not match):
        if height!=self.height or width!=self.width:
            if self.a1!=NULL or self.a2!=NULL or self.distances!=NULL:
                log("new image size: %ix%i (was %ix%i), clearing reference checksums", width, height, self.width, self.height)
                self.free()
            self.width = width
            self.height = height
        #this is a new picture, shift a2 into a1 if we have it:
        if self.a1:
            free(self.a1)
            self.a1 = NULL
        if self.a2:
            self.a1 = self.a2
            self.a2 = NULL
        cdef size_t row_len = width*bpp
        #allocate new checksum array:
        assert self.a2==NULL
        cdef size_t asize = height*(sizeof(uint64_t))
        self.a2 = <uint64_t*> memalign(asize)
        assert self.a2!=NULL, "checksum memory allocation failed"
        #checksum each line of the pixel array:
        cdef Py_ssize_t min_buf_len = rowstride*height
        cdef uint64_t *a2 = self.a2
        cdef uint16_t i
        cdef uint8_t *buf
        with buffer_context(pixels) as bc:
            buf = <uint8_t*> (<uintptr_t> int(bc))
            assert len(bc)>=min_buf_len, "buffer length=%i is too small for %ix%i with rowstride %i, should be %i" % (
                    len(bc), width, height, rowstride, min_buf_len)
            assert row_len<=rowstride, "invalid row length: %ix%i=%i but rowstride is %i" % (width, bpp, width*bpp, rowstride)
            with nogil:
                for i in range(height):
                    a2[i] = xxh3(buf, row_len)
                    buf += rowstride


    def calculate(self, uint16_t max_distance=1000) -> None:
        """
            Find all the scroll distances
            that would move lines from a1 to a2.
            The same lines may be accounted for multiple times.
            The result is stored in the "distances" array.
        """
        if DEBUG:
            log("calculate(%i) a1=%#x, a2=%#x, distances=%#x", max_distance, <uintptr_t> self.a1, <uintptr_t> self.a2, <uintptr_t> self.distances)
        if self.a1==NULL or self.a2==NULL:
            return
        cdef uint64_t *a1 = self.a1
        cdef uint64_t *a2 = self.a2
        cdef uint16_t l = self.height
        cdef uint16_t y1, y2
        cdef uint16_t miny, maxy
        cdef uint64_t a2v
        if self.distances==NULL:
            self.distances = <uint16_t*> memalign(2*l*sizeof(uint16_t))
            assert self.distances!=NULL, "distance memory allocation failed"
        cdef uint32_t matches = 0
        with nogil:
            memset(self.distances, 0, 2*l*sizeof(uint16_t))
            for y2 in range(l):
                #miny = max(0, y2-max_distance):
                if y2>max_distance:
                    miny = y2-max_distance
                else:
                    miny = 0
                #maxy = min(l, y2+max_distance)
                if y2+max_distance<l:
                    maxy = y2+max_distance
                else:
                    maxy = l
                a2v = a2[y2]
                if a2v==0:
                    continue
                for y1 in range(miny, maxy):
                    if a1[y1]==a2v:
                        #distance = y1-y2
                        self.distances[l-(y1-y2)] += 1
                        matches += 1
        if DEBUG:
            log("ScrollDistance: height=%i, calculate:", l)
            log(" a1=%s", da(self.a1, l))
            log(" a2=%s", da(self.a2, l))
            log(" %i matches, distances=%s", matches, dd(self.distances, l*2))

    def get_scroll_values(self, uint16_t min_hits=2) -> Tuple[Dict, Dict]:
        """
            Return two dictionaries that describe how to go from a1 to a2.
            * scrolls dictionary contains scroll definitions
            * non-scrolls dictionary is everything else (that will need to be repainted)
        """
        DEF MAX_MATCHES = 20
        if self.a1==NULL or self.a2==NULL:
            return None
        cdef uint16_t m_arr[MAX_MATCHES]    #number of hits
        cdef int16_t s_arr[MAX_MATCHES]     #scroll distance
        cdef int16_t i
        cdef uint8_t j
        cdef int16_t low = 0                #the lowest match value
        cdef int16_t matches
        cdef uint16_t* distances = self.distances
        cdef uint16_t l = self.height
        cdef size_t asize = l*sizeof(uint8_t)
        #use a temporary buffer to track the lines we have already dealt with:
        cdef uint8_t *line_state = <uint8_t*> malloc(asize)
        assert line_state!=NULL, "state map memory allocation failed"
        #find the best values (highest match count):
        with nogil:
            memset(line_state, 0, asize)
            memset(m_arr, 0, MAX_MATCHES*sizeof(uint16_t))
            memset(s_arr, 0, MAX_MATCHES*sizeof(int16_t))
            for i in range(2*l):
                matches = distances[i]
                if matches>low and matches>min_hits:
                    #add this candidate match to the arrays:
                    #find the lowest score index and replace it:
                    for j in range(MAX_MATCHES):
                        if m_arr[j]==low:
                            break
                    m_arr[j] = matches
                    s_arr[j] = i-l
                    #find the new lowest value we have:
                    low = matches
                    for j in range(MAX_MATCHES):
                        if m_arr[j]<low:
                            low = m_arr[j]
                            if low==0:
                                break
        #first collect the list of distances:
        #(there can be more than one distance value for each match count):
        scroll_hits = {}
        for i in range(MAX_MATCHES):
            if m_arr[i]>min_hits:
                scroll_hits.setdefault(m_arr[i], []).append(s_arr[i])
        if DEBUG:
            log("scroll hits=%s", dict(reversed(sorted(scroll_hits.items()))))
        #return a dict with the scroll distance as key,
        #and the list of matching lines as value:
        # {line-start : count, ..}
        cdef uint16_t start = 0, count = 0
        try:
            scrolls = {}
            #starting with the highest matches
            for i in reversed(sorted(scroll_hits.keys())):
                v = scroll_hits[i]
                for scroll in v:
                    #find matching lines:
                    line_defs = self.match_distance(line_state, scroll, MIN_LINE_COUNT)
                    if line_defs:
                        scrolls[scroll] = line_defs
            #same for the unmatched lines:
            #all the lines in tmp which have not been set by match_distance()
            line_defs = {}
            for i in range(l):
                if line_state[i]==0:
                    if count==0:
                        start = i
                    count += 1
                elif count>0:
                    line_defs[start] = count
                    count = 0
            if count>0:
                line_defs[start] = count
        finally:
            free(line_state)
        return scrolls, line_defs

    cdef match_distance(self, uint8_t *line_state, int16_t distance, const uint8_t min_line_count):
        """
            find the lines that match the given scroll distance,
            return a dictionary with the starting line as key
            and the number of matching lines as value
        """
        cdef uint64_t *a1 = self.a1
        cdef uint64_t *a2 = self.a2
        cdef uint64_t v
        assert abs(distance)<=self.height, "invalid distance %i for size %i" % (distance, self.height)
        cdef uint16_t rstart = 0
        cdef uint16_t rend = self.height-distance
        if distance<0:
            rstart = -distance
            rend = self.height
        cdef uint16_t i1, i2, start = 0, count = 0
        line_defs = {}
        for i1 in range(rstart, rend):
            i2 = i1+distance
            v = a1[i1]
            #if DEBUG:
            #    log("%i: a1=%i / a2=%i", i, a1[i], a2[i2])
            if v==a2[i2] and v!=0:
                #if DEBUG:
                #    log("match at %i: %i", i, a1[i])
                if count==0:
                    if line_state[i2]:
                        #this line has been matched already,
                        #we don't need to start here
                        continue
                    start = i1
                count += 1
            elif count>0:
                #we had a match
                if count>min_line_count:
                    line_defs[start] = count
                count = 0
        if count>min_line_count:
            #last few lines ended as a match:
            line_defs[start] = count
        #clear the ones we have matched:
        for start, count in line_defs.items():
            for i1 in range(count):
                line_state[start+distance+i1] = 1
        #if DEBUG:
        #    log("match_distance(%i)=%s", distance, line_defs)
        return line_defs


    def invalidate(self, int16_t x, int16_t y, uint16_t w, uint16_t h) -> None:
        if self.a2==NULL:
            #nothing to invalidate!
            return
        #do they intersect?
        rect = rectangle(self.x, self.y, self.width, self.height)
        inter = rect.intersection(x, y, w, h)
        if not inter:
            return
        #remove any lines that have been updated
        #by zeroing out their checksums:
        assert inter.height<=self.height
        assert inter.y>=rect.y and inter.y+inter.height<=rect.y+rect.height
        #the array indexes are relative to rect.y:
        cdef int start_y = inter.y-rect.y
        cdef int i
        for i in range(start_y, start_y+inter.height):
            self.a2[i] = 0
        cdef uint16_t nonzero = 0
        for i in range(self.height):
            if self.a2[i]!=0:
                nonzero += 1
        log("invalidated %i lines checksums from intersection of scroll area %s and rectangle %s, remains %i", inter.height, rect, (x, y, w, h), nonzero)
        #if more than half has already been invalidated, drop it completely:
        if nonzero<=rect.height//2:
            log("invalidating whole scroll data as only %i of it remains valid", 100*nonzero//rect.height)
            self.free()

    def get_best_match(self) -> Tuple[int, int]:
        if self.a1==NULL or self.a2==NULL:
            return 0, 0
        cdef uint16_t max_hits = 0
        cdef int d = 0
        cdef unsigned int i
        cdef uint16_t r = 2*self.height
        for i in range(r):
            if self.distances[i]>max_hits:
                max_hits = self.distances[i]
                d = i-self.height
        return d, max_hits

    def __dealloc__(self):
        self.free()

    def free(self) -> None:
        cdef void* ptr = <void*> self.distances
        if ptr:
            self.distances = NULL
            free(ptr)
        ptr = <void*> self.a1
        if ptr:
            self.a1 = NULL
            free(ptr)
        ptr = <void*> self.a2
        if ptr:
            self.a2 = NULL
            free(ptr)



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False
from typing import Tuple
from collections.abc import Sequence

from xpra.log import Logger
log = Logger("decoder", "spng")

from xpra.codecs.image import ImageWrapper
from xpra.net.compression import Compressed
from xpra.codecs.debug import may_save_image
from xpra.codecs.argb.argb import unpremultiply_argb
from xpra.codecs.spng.spng cimport (
    SPNG_VERSION_MAJOR, SPNG_VERSION_MINOR, SPNG_VERSION_PATCH,
    SPNG_CTX_ENCODER,
    SPNG_INTERLACE_NONE,
    SPNG_ENCODE_FINALIZE,
    SPNG_FILTER_CHOICE, SPNG_FILTER_CHOICE_NONE, SPNG_FILTER_CHOICE_SUB,
    SPNG_ENCODE_TO_BUFFER,
    SPNG_IMG_COMPRESSION_LEVEL,
    SPNG_IMG_MEM_LEVEL,
    SPNG_IMG_COMPRESSION_STRATEGY,
    SPNG_TEXT_WINDOW_BITS,
    SPNG_FMT_PNG,
    SPNG_COLOR_TYPE_GRAYSCALE, SPNG_COLOR_TYPE_TRUECOLOR, SPNG_COLOR_TYPE_GRAYSCALE_ALPHA, SPNG_COLOR_TYPE_TRUECOLOR_ALPHA,
    spng_ctx, spng_ihdr, spng_strerror,
    spng_ctx_new, spng_ctx_free,
    spng_set_option, spng_set_ihdr,
    spng_encode_image, spng_get_png_buffer,
    spng_format,
)
from libc.stdint cimport uintptr_t, uint32_t, uint8_t
from xpra.buffers.membuf cimport makebuf, MemBuf, buffer_context  # pylint: disable=syntax-error


cdef extern from "zconf.h":
    int MAX_MEM_LEVEL

cdef extern from "zlib.h":
    int Z_HUFFMAN_ONLY


def get_version() -> Tuple[int, int, int]:
    return (SPNG_VERSION_MAJOR, SPNG_VERSION_MINOR, SPNG_VERSION_PATCH)


def get_type() -> str:
    return "spng"


def get_encodings() -> Sequence[str]:
    return ("png", "png/L")


def get_error_str(int r) -> str:
    b = spng_strerror(r)
    return b.decode()


def check_error(int r, msg) -> int:
    if r:
        log_error(r, msg)
    return r


def log_error(int r, msg) -> None:
    log.error("Error: %s", msg)
    log.error(" code %i: %s", r, get_error_str(r))


INPUT_FORMATS = "RGBA", "RGB"


def encode(coding: str, image: ImageWrapper, options=None) -> Tuple:
    assert coding in ("png", "png/L")
    options = options or {}
    cdef int grayscale = options.get("grayscale", 0) or coding=="png/L"
    cdef int speed = options.get("speed", 50)
    cdef int width = image.get_width()
    cdef int height = image.get_height()
    cdef int rowstride = image.get_rowstride()
    cdef int scaled_width = options.get("scaled-width", width)
    cdef int scaled_height = options.get("scaled-height", height)
    cdef char resize = scaled_width!=width or scaled_height!=height

    rgb_format = image.get_pixel_format()
    alpha = options.get("alpha", rgb_format.find("A")>=0)
    if rgb_format not in INPUT_FORMATS or (resize and len(rgb_format)!=4) or rowstride!=width*len(rgb_format) or grayscale or alpha:
        #best to restride before byte-swapping to trim extra unused data:
        if rowstride!=width*len(rgb_format):
            image.restride(width*len(rgb_format))
        input_formats = INPUT_FORMATS
        if alpha:
            image.set_pixels(unpremultiply_argb(image.get_pixels()))
        if grayscale:
            input_formats = ("BGRX", "BGRA", "BGR", "RGB")
        if rgb_format not in input_formats:
            from xpra.codecs.argb.argb import argb_swap
            if not argb_swap(image, input_formats, supports_transparency=alpha):
                log("spng: argb_swap failed to convert %s to a suitable format: %s" % (
                    rgb_format, input_formats))
            else:
                log("spng converted %s to %s", rgb_format, image)
                rgb_format = image.get_pixel_format()
        rowstride = image.get_rowstride()

    if resize:
        from xpra.codecs.argb.scale import scale_image
        image = scale_image(image, scaled_width, scaled_height)
        log("spng scaled image: %s", image)

    pixels = image.get_pixels()
    if grayscale:
        from xpra.codecs.argb.argb import bgrx_to_l, bgra_to_la, bgr_to_l, rgb_to_l
        if alpha:
            pixels = bgra_to_la(pixels)
            rgb_format = "LA"
        elif rgb_format=="BGRX":
            pixels = bgrx_to_l(pixels)
            rgb_format = "L"
        elif rgb_format=="BGR":
            pixels = bgr_to_l(pixels)
            rgb_format = "L"
        elif rgb_format=="RGB":
            pixels = rgb_to_l(pixels)
            rgb_format = "L"
        else:
            raise ValueError(f"invalid rgb pixel format {rgb_format!r}")

    cdef spng_ctx *ctx = spng_ctx_new(SPNG_CTX_ENCODER)
    if ctx==NULL:
        raise RuntimeError("failed to instantiate an spng context")

    cdef spng_ihdr ihdr
    ihdr.width = scaled_width
    ihdr.height = scaled_height
    ihdr.bit_depth = 8
    if rgb_format=="L":
        ihdr.color_type = SPNG_COLOR_TYPE_GRAYSCALE
    elif rgb_format=="LA":
        ihdr.color_type = SPNG_COLOR_TYPE_GRAYSCALE_ALPHA
    elif rgb_format=="RGBA":
        ihdr.color_type = SPNG_COLOR_TYPE_TRUECOLOR_ALPHA
    elif rgb_format=="RGB":
        ihdr.color_type = SPNG_COLOR_TYPE_TRUECOLOR
    else:
        raise ValueError(f"unsupported input pixel format {rgb_format!r}")

    ihdr.compression_method = 0
    ihdr.filter_method = 0
    ihdr.interlace_method = SPNG_INTERLACE_NONE
    if check_error(spng_set_ihdr(ctx, &ihdr),
                   "failed to set encode-to-buffer option"):
        spng_ctx_free(ctx)
        return None

    cdef int clevel = 1
    if check_error(spng_set_option(ctx, SPNG_IMG_COMPRESSION_LEVEL, clevel),
                   "failed to set compression level"):
        spng_ctx_free(ctx)
        return ()
    if check_error(spng_set_option(ctx, SPNG_TEXT_WINDOW_BITS, 15),
                   "failed to set window bits"):
        spng_ctx_free(ctx)
        return ()
    if check_error(spng_set_option(ctx, SPNG_IMG_COMPRESSION_STRATEGY, Z_HUFFMAN_ONLY),
                   "failed to set compression strategy"):
        spng_ctx_free(ctx)
        return ()
    if check_error(spng_set_option(ctx, SPNG_IMG_MEM_LEVEL, MAX_MEM_LEVEL),
                   "failed to set mem level"):
        spng_ctx_free(ctx)
        return ()
    cdef int filter = SPNG_FILTER_CHOICE_NONE
    if speed<30:
        filter |= SPNG_FILTER_CHOICE_SUB
    if check_error(spng_set_option(ctx, SPNG_FILTER_CHOICE, filter),
                   "failed to set filter choice"):
        spng_ctx_free(ctx)
        return ()

    if check_error(spng_set_option(ctx, SPNG_ENCODE_TO_BUFFER, 1),
                   "failed to set encode-to-buffer option"):
        spng_ctx_free(ctx)
        return ()

    cdef spng_format fmt = SPNG_FMT_PNG
    cdef int flags = SPNG_ENCODE_FINALIZE
    cdef size_t data_len = 0
    cdef uintptr_t data_ptr
    cdef int r = 0
    assert len(pixels)>=ihdr.width*ihdr.height*len(rgb_format), \
        "pixel buffer is too small, expected %i bytes but got %i for %ix%i '%s'" % (
        ihdr.width*ihdr.height*4, len(pixels), ihdr.width, ihdr.height, rgb_format)
    with buffer_context(pixels) as bc:
        data_ptr = <uintptr_t> int(bc)
        data_len = len(bc)
        #log("spng encode buffer %#x, len=%#x", <uintptr_t> data_ptr, data_len)
        with nogil:
            r = spng_encode_image(ctx, <const void*> data_ptr, data_len, fmt, flags)
    if check_error(r, "failed to encode image"):
        log.error(" %i bytes of %s pixel data", data_len, rgb_format)
        log.error(" for %s", image)
        spng_ctx_free(ctx)
        return ()

    cdef int error
    cdef size_t png_len
    cdef void *png_data = spng_get_png_buffer(ctx, &png_len, &error)
    if check_error(error,
                   "failed get png buffer"):
        spng_ctx_free(ctx)
        return ()
    if png_data==NULL:
        log.error("Error: spng buffer is NULL")
        spng_ctx_free(ctx)
        return ()

    cdef membuf = makebuf(png_data, png_len)
    spng_ctx_free(ctx)
    cdata = memoryview(membuf).toreadonly()
    may_save_image("png", cdata)
    return coding, Compressed(coding, cdata), {}, width, height, 0, len(rgb_format)*8


def selftest(full=False) -> None:
    log("spng version %s selftest" % (get_version(),))
    from xpra.codecs.checks import make_test_image
    for rgb_format in ("RGBA", "RGB", "BGRA", "BGRX"):
        image = make_test_image(rgb_format, 1024, 768)
        assert encode("png", image), "failed to encode %s" % image



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False

import struct
from typing import Tuple

from xpra.common import SizedBuffer
from libc.stdint cimport uintptr_t
from xpra.buffers.membuf cimport MemBuf, getbuf

from xpra.log import Logger
log = Logger("lz4")


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "lz4_compat.h":
    int LZ4_versionNumber()
    int LZ4_MAX_INPUT_SIZE

    ctypedef struct LZ4_stream_t:
        pass

    int LZ4_compressBound(int inputSize)
    void LZ4_resetStream_fast(LZ4_stream_t* stream)
    int LZ4_compress_fast_continue(LZ4_stream_t* stream,
                                   const char* src, char* dst,
                                   int srcSize, int dstCapacity, int acceleration) nogil

    int LZ4_decompress_safe(const char* src, char* dst, int compressedSize, int dstCapacity) nogil


def get_version() -> Tuple[int, int, int]:
    cdef int v = LZ4_versionNumber()
    return v//10000, (v//100) % 100, v % 100


cdef class compressor:
    cdef LZ4_stream_t state

    def __init__(self):
        LZ4_resetStream_fast(&self.state)

    def bound(self, int size) -> int:
        return LZ4_compressBound(size)

    def compress(self, data: SizedBuffer, int acceleration=1, int max_size=0, int store_size=True) -> memoryview:
        cdef Py_buffer in_buf
        if PyObject_GetBuffer(data, &in_buf, PyBUF_ANY_CONTIGUOUS):
            raise ValueError("failed to read data from %s" % type(data))
        if in_buf.len > LZ4_MAX_INPUT_SIZE:
            PyBuffer_Release(&in_buf)
            log("input is too large")
            raise ValueError("input is too large")
        if max_size <= 0:
            max_size = LZ4_compressBound(in_buf.len)
        cdef int size_header = 0
        if store_size:
            size_header = 4
        cdef MemBuf out_buf = getbuf(size_header + max_size, False)
        mem = memoryview(out_buf)
        if store_size:
            struct.pack_into(b"@I", mem, 0, in_buf.len)
        cdef const char *in_ptr = <const char *> in_buf.buf
        cdef char *out_ptr = <char *> ((<uintptr_t> out_buf.get_mem())+size_header)
        cdef int r
        with nogil:
            r = LZ4_compress_fast_continue(&self.state, in_ptr, out_ptr, in_buf.len, max_size-size_header, acceleration)
        if r <= 0:
            msg = f"LZ4_compress_fast_continue failed for input size {in_buf.len} and output buffer size {max_size}"
            log(msg)
            raise ValueError(msg)
        PyBuffer_Release(&in_buf)
        return (mem[:(size_header+r)]).toreadonly()


def compress(data: SizedBuffer, acceleration=1) -> memoryview:
    c = compressor()
    return c.compress(data, acceleration)


def decompress(data: SizedBuffer, int max_size=0, int size=0) -> memoryview:
    cdef int size_header = 0
    if size == 0:
        size = struct.unpack_from(b"@I", data[:4])[0]
        size_header = 4
    if max_size > 0 and size > max_size:
        raise ValueError("data would overflow max-size %i" % max_size)
    if size > LZ4_MAX_INPUT_SIZE:
        raise ValueError("data would overflow lz4 max input size %i" % LZ4_MAX_INPUT_SIZE)
    cdef Py_buffer in_buf
    if PyObject_GetBuffer(data, &in_buf, PyBUF_ANY_CONTIGUOUS):
        raise ValueError("failed to read data from %s" % type(data))
    cdef MemBuf out_buf = getbuf(size, 0)
    cdef char *in_ptr = <char*> ((<uintptr_t> in_buf.buf) + size_header)
    cdef char *out_ptr = <char *> out_buf.get_mem()
    cdef int l = <int> in_buf.len
    cdef int r
    with nogil:
        r = LZ4_decompress_safe(in_ptr, out_ptr, l-size_header, size)
    PyBuffer_Release(&in_buf)
    if r <= 0:
        msg = f"LZ4_decompress_safe failed for input size {in_buf.len}"
        log(msg)
        raise ValueError(msg)
    return memoryview(out_buf)[:r]



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import struct
from time import monotonic
from socket import error as socket_error
from collections.abc import Callable, Sequence
from queue import Queue

from xpra.os_util import gi_import
from xpra.util.str_fn import repr_ellipsized, hexstr
from xpra.util.env import envint
from xpra.util.thread import make_thread, start_thread
from xpra.util.stats import std_unit
from xpra.net.protocol.socket_handler import force_flush_queue, exit_queue
from xpra.net.protocol.constants import INVALID, CONNECTION_LOST
from xpra.net.common import ConnectionClosedException  # @UndefinedVariable (pydev false positive)
from xpra.net.bytestreams import ABORT
from xpra.net.rfb.const import RFBClientMessage, CLIENT_PACKET_TYPE_STR, PACKET_STRUCT
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("network", "protocol", "rfb")

RFB_LOG = os.environ.get("XPRA_RFB_LOG")
READ_BUFFER_SIZE = envint("XPRA_READ_BUFFER_SIZE", 65536)

PROTOCOL_VERSION = (3, 8)


class RFBProtocol:
    TYPE = "rfb"

    def __init__(self, conn, process_packet_cb, data=b""):
        if not conn:
            raise ValueError(f"no connection: {conn}")
        self.start_time = monotonic()
        self._conn = conn
        self._process_packet_cb = process_packet_cb
        self._write_queue = Queue()
        self._buffer = data
        self._challenge = None
        self.share = False
        # counters:
        self.input_packetcount = 0
        self.input_raw_packetcount = 0
        self.output_packetcount = 0
        self.output_raw_packetcount = 0
        self._closed = False
        self._packet_parser = self._parse_protocol_handshake
        self._write_thread = None
        self._read_thread = make_thread(self._read_thread_loop, "read", daemon=True)
        self.log = None
        if RFB_LOG:
            # pylint: disable=consider-using-with
            self.log = open(RFB_LOG, "w", encoding="utf8")

    def is_closed(self) -> bool:
        return self._closed

    # noinspection PyMethodMayBeStatic
    def is_sending_encrypted(self) -> bool:
        return False

    def send_protocol_handshake(self) -> None:
        self.send(b"RFB 003.008\n")

    # noinspection PyMethodMayBeStatic
    def _parse_invalid(self, packet) -> int:
        return len(packet)

    def _parse_protocol_handshake(self, packet) -> int:
        log(f"parse_protocol_handshake({packet})")
        if len(packet) < 12:
            return 0
        if not packet.startswith(b'RFB '):
            self.invalid_header(self, packet, "invalid RFB protocol handshake packet header")
            return 0
        # ie: packet==b'RFB 003.008\n'
        protocol_version = tuple(int(x) for x in packet[4:11].split(b"."))
        if protocol_version != PROTOCOL_VERSION:
            msg = b"unsupported protocol version"
            log.error(f"Error: {msg!r}")
            self.send(struct.pack(b"!BI", 0, len(msg)) + msg)
            self.invalid(msg, packet)
            return 0
        self.handshake_complete()
        return 12

    def handshake_complete(self):
        raise NotImplementedError

    def _parse_security_handshake(self, packet):
        raise NotImplementedError

    def _parse_challenge(self, response):
        raise NotImplementedError

    def _parse_security_result(self, packet):
        raise NotImplementedError

    def _parse_rfb(self, packet) -> int:
        try:
            ptype = ord(packet[0])
        except TypeError:
            ptype = packet[0]
        packet_type = CLIENT_PACKET_TYPE_STR.get(ptype)
        if not packet_type:
            self.invalid(f"unknown RFB packet type: {ptype:x}", packet)
            return 0
        s = PACKET_STRUCT.get(ptype)  # ie: Struct("!BBBB")
        if not s:
            self.invalid(f"RFB packet type {packet_type!r} is not supported", packet)
            return 0
        if len(packet) < s.size:
            return 0
        size = s.size
        values = list(s.unpack(packet[:size]))
        values[0] = packet_type
        # some packets require parsing extra data:
        if ptype == RFBClientMessage.SetEncodings:
            N = values[2]
            estruct = struct.Struct(b"!" + b"i" * N)
            size += estruct.size
            if len(packet) < size:
                return 0
            encodings = estruct.unpack(packet[s.size:size])
            values.append(encodings)
        elif ptype == RFBClientMessage.ClientCutText:
            count = values[4]
            size += count
            if len(packet) < size:
                return 0
            text = packet[s.size:size]
            values.append(text)
        self.input_packetcount += 1
        log(f"RFB packet: {packet_type}: {values[1:]}")
        # now trigger the callback:
        self._process_packet_cb(self, values)
        # return part of packet not consumed:
        return size

    def __repr__(self):
        return f"RFBProtocol({self._conn})"

    def get_threads(self) -> tuple:
        return tuple(x for x in (self._write_thread, self._read_thread) if x is not None)

    def get_info(self, *_args) -> dict[str, Sequence[int] | dict[str, bool]]:
        info: dict[str, tuple | dict] = {
            "protocol": PROTOCOL_VERSION,
            "thread": dict((thread.name, thread.is_alive()) for thread in self.get_threads()),
        }
        return info

    def start(self) -> None:
        def start_network_read_thread():
            if not self._closed:
                self._read_thread.start()

        GLib.idle_add(start_network_read_thread)

    def send_disconnect(self, *_args, **_kwargs) -> None:
        # no such packet in RFB, just close
        self.close()

    def queue_size(self) -> int:
        return self._write_queue.qsize()

    def send(self, packet) -> None:
        if self._closed:
            log("connection is closed already, not sending packet")
            return
        if log.is_debug_enabled():
            size = len(packet)
            lstr = str(size) if size <= 16 else std_unit(size)
            log(f"send({lstr} bytes: %s..)", hexstr(packet[:16]))
        if self.log:
            self.log.write(f"send: {hexstr(packet)}\n")
        if self._write_thread is None:
            self.start_write_thread()
        self._write_queue.put(packet)

    def start_write_thread(self) -> None:
        log("rfb: starting write thread")
        self._write_thread = start_thread(self._write_thread_loop, "write", daemon=True)

    def _io_thread_loop(self, name: str, callback: Callable) -> None:
        try:
            log(f"io_thread_loop({name}, {callback}) loop starting")
            while not self._closed and callback():
                "wait for an exit condition"
            log(f"io_thread_loop({name}, {callback}) loop ended, closed={self._closed}")
        except ConnectionClosedException:
            log(f"{self._conn} closed", exc_info=True)
            if not self._closed:
                # ConnectionClosedException means the warning has been logged already
                self._connection_lost(f"{name} connection {self._conn} closed")
        except (OSError, socket_error) as e:
            if not self._closed:
                self._internal_error(f"{name} connection {self._conn} reset", e, exc_info=e.args[0] not in ABORT)
        except Exception as e:
            # can happen during close(), in which case we just ignore:
            if not self._closed:
                log.error(f"Error: {name} on {self._conn} failed: {type(e)}", exc_info=True)
                self.close()

    def _write_thread_loop(self) -> None:
        self._io_thread_loop("write", self._write)

    def _write(self) -> bool:
        buf = self._write_queue.get()
        # Used to signal that we should exit:
        if buf is None:
            log("write thread: empty marker, exiting")
            self.close()
            return False
        con = self._conn
        if not con:
            return False
        while buf and not self._closed:
            written = con.write(buf)
            if written:
                buf = buf[written:]
                self.output_raw_packetcount += 1
        self.output_packetcount += 1
        return True

    def _read_thread_loop(self) -> None:
        self._io_thread_loop("read", self._read)

    def _read(self) -> bool:
        c = self._conn
        if not c:
            return False
        buf = c.read(READ_BUFFER_SIZE)
        # log("read()=%i bytes (%s)", len(buf or b""), type(buf))
        if not buf:
            log("read thread: eof")
            # give time to the parse thread to call close itself,
            # so it has time to parse and process the last packet received
            GLib.timeout_add(1000, self.close)
            return False
        if self.log:
            self.log.write(f"receive: {hexstr(buf)}\n")
        self.input_raw_packetcount += 1
        self._buffer += buf
        # log("calling %s(%s)", self._packet_parser, repr_ellipsized(self._buffer))
        while self._buffer:
            consumed = self._packet_parser(self._buffer)
            if consumed == 0:
                break
            self._buffer = self._buffer[consumed:]
        return True

    def _internal_error(self, message="", exc=None, exc_info=False) -> None:
        # log exception info with last log message
        if self._closed:
            return
        ei = exc_info
        if exc:
            ei = None  # log it separately below
        log.error(f"Error: {message}", exc_info=ei)
        if exc:
            log.error(f" {exc}", exc_info=exc_info)
        GLib.idle_add(self._connection_lost, message)

    def _connection_lost(self, message="", exc_info=False) -> bool:
        log(f"connection lost: {message}", exc_info=exc_info)
        self.close()
        return False

    def invalid(self, msg, data) -> None:
        log("invalid(%s, %r)", msg, data)
        self._packet_parser = self._parse_invalid
        GLib.idle_add(self._process_packet_cb, self, [INVALID, msg, data])
        # Then hang up:
        GLib.timeout_add(1000, self._connection_lost, msg)

    # delegates to invalid_header()
    def invalid_header(self, proto, data, msg="") -> None:
        log("invalid_header%s", (proto, data, msg))
        self._invalid_header(proto, data, msg)

    def _invalid_header(self, _proto, data, msg="invalid packet header") -> None:
        self._packet_parser = self._parse_invalid
        err = f"{msg}: {hexstr(data[:8])!r}"
        if len(data) > 1:
            err += f" read buffer={repr_ellipsized(data)} ({len(data)} bytes)"
        self.invalid(err, data)

    def gibberish(self, msg, data) -> None:
        log(f"gibberish({msg}, {data!r})")
        self.close()

    def close(self) -> None:
        c = self._conn
        log(f"RFBProtocol.close() closed={self._closed}, connection={self._conn}")
        if self._closed:
            return
        self._closed = True
        # GLib.idle_add(self._process_packet_cb, self, [CONNECTION_LOST])
        if c:
            try:
                log(f"RFBProtocol.close() calling {c.close}")
                c.close()
            except OSError:
                log.error(f"Error closing {c}", exc_info=True)
            self._conn = None
        self.terminate_queue_threads()
        self._process_packet_cb(self, [CONNECTION_LOST])
        GLib.idle_add(self.clean)
        log_file = self.log
        if log_file:
            self.log = None
            log_file.close()
        log("RFBProtocol.close() done")

    def clean(self) -> None:
        # clear all references to ensure we can get garbage collected quickly:
        self._write_thread = None
        self._read_thread = None
        self._process_packet_cb = None

    def terminate_queue_threads(self) -> None:
        log("terminate_queue_threads()")
        # make all the queue based threads exit by adding the empty marker:
        owq = self._write_queue
        self._write_queue = exit_queue()
        force_flush_queue(owq)



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from time import monotonic
from typing import Any

from xpra.server.core import ServerCore
from xpra.server.background_worker import add_work_item
from xpra.common import FULL_INFO, noop, ConnectionMessage
from xpra.net.common import PacketType, PacketElement
from xpra.scripts.config import str_to_bool
from xpra.os_util import WIN32, gi_import
from xpra.util.objects import typedict, merge_dicts
from xpra.util.str_fn import csv, Ellipsizer
from xpra.util.env import envbool
from xpra.net.bytestreams import set_socket_timeout
from xpra.server import features, ServerExitMode
from xpra.server.factory import get_server_base_classes
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("server")
netlog = Logger("network")
authlog = Logger("auth")

SERVER_BASES = get_server_base_classes()
ServerBaseClass = type("ServerBaseClass", SERVER_BASES, {})
log("ServerBaseClass%s", SERVER_BASES)

CLIENT_CAN_SHUTDOWN = envbool("XPRA_CLIENT_CAN_SHUTDOWN", True)
MDNS_CLIENT_COUNT = envbool("XPRA_MDNS_CLIENT_COUNT", True)


class ServerBase(ServerBaseClass):
    """
    This is the base class for seamless and desktop servers. (not proxy servers)
    It provides all the generic functions but is not tied
    to a specific backend (X11 or otherwise).
    See GTKServerBase/X11ServerBase and other platform specific subclasses.
    """

    def __init__(self):
        for c in SERVER_BASES:
            c.__init__(self)
        log("ServerBase.__init__()")

        self._server_sources: dict = {}
        self.client_properties: dict[int, dict] = {}
        self.ui_driver = None
        self.sharing: bool | None = None
        self.lock: bool | None = None
        self.idle_timeout: int = 0
        self.client_shutdown: bool = CLIENT_CAN_SHUTDOWN

    def suspend_event(self, args) -> None:
        ServerCore.suspend_event(self, args)
        self.server_event("suspend")
        for s in self._server_sources.values():
            if hasattr(s, "go_idle"):
                s.go_idle()

    def resume_event(self, args) -> None:
        ServerCore.resume_event(self, args)
        self.server_event("resume")
        for s in self._server_sources.values():
            if hasattr(s, "no_idle"):
                s.no_idle()

    def server_event(self, event_type: str, *args: PacketElement) -> None:
        for s in self._server_sources.values():
            s.send_server_event(event_type, *args)
        # the bus mixin is optional:
        dbus_server = getattr(self, "dbus_server", None)
        if dbus_server:
            dbus_server.Event(event_type, [str(x) for x in args[1:]])

    def get_server_source(self, proto):
        return self._server_sources.get(proto)

    def init(self, opts) -> None:
        # from now on, use the logger for parsing errors:
        from xpra.scripts import config  # pylint: disable=import-outside-toplevel
        config.warn = log.warn
        for c in SERVER_BASES:
            start = monotonic()
            c.init(self, opts)
            end = monotonic()
            log("%3ims in %s.init", 1000 * (end - start), c)
        self.sharing = opts.sharing
        self.lock = opts.lock
        self.idle_timeout = opts.idle_timeout

    def setup(self) -> None:
        log("starting component init")
        for c in SERVER_BASES:
            start = monotonic()
            c.setup(self)
            end = monotonic()
            log("%3ims in %s.setup", 1000 * (end - start), c)

    def threaded_init(self) -> None:
        log("threaded_init() serverbase start")
        for c in SERVER_BASES:
            with log.trap_error("Error during threaded setup of %s", c):
                c.threaded_setup(self)
        log("threaded_init() serverbase end")
        super().call_init_thread_callbacks()

    def server_is_ready(self) -> None:
        ServerCore.server_is_ready(self)
        self.server_event("ready")

    def do_cleanup(self) -> None:
        self.server_event("exit")
        self.wait_for_threaded_init()
        log("do_cleanup() calling on %s", SERVER_BASES)
        for c in reversed(SERVER_BASES):
            if c != ServerCore:
                log("%s", c.cleanup)
                c.cleanup(self)

    def late_cleanup(self, stop=True) -> None:
        for c in reversed(SERVER_BASES):
            log("%s", c.late_cleanup)
            c.late_cleanup(self, stop)

    ######################################################################
    # override http scripts to expose just the current session / display
    def get_displays(self) -> dict[str, Any]:
        from xpra.scripts.main import get_displays  # pylint: disable=import-outside-toplevel
        return get_displays(self.dotxpra, display_names=(os.environ.get("DISPLAY"),))

    def get_xpra_sessions(self) -> dict[str, Any]:
        from xpra.scripts.main import get_xpra_sessions  # pylint: disable=import-outside-toplevel
        return get_xpra_sessions(self.dotxpra, matching_display=os.environ.get("DISPLAY"))

    ######################################################################
    # shutdown / exit commands:
    def _process_exit_server(self, _proto, packet: PacketType = ("exit-server", )) -> None:
        reason: ConnectionMessage | str = ConnectionMessage.SERVER_EXIT
        message = "Exiting in response to client request"
        if len(packet) > 1:
            reason = str(packet[1])
            message += ": " + reason
        log.info(message)
        self.cleanup_all_protocols(reason=reason)
        GLib.timeout_add(500, self.clean_quit, ServerExitMode.EXIT)

    def _process_shutdown_server(self, _proto, _packet: PacketType = ("shutdown-server", )) -> None:
        if not self.client_shutdown:
            log.warn("Warning: ignoring shutdown request")
            return
        log.info("Shutting down in response to client request")
        self.cleanup_all_protocols(reason=ConnectionMessage.SERVER_SHUTDOWN)
        GLib.timeout_add(500, self.clean_quit)

    def get_mdns_info(self) -> dict[str, Any]:
        mdns_info = ServerCore.get_mdns_info(self)
        if MDNS_CLIENT_COUNT:
            mdns_info["clients"] = len(self._server_sources)
        return mdns_info

    ######################################################################
    # handle new connections:
    def handle_sharing(self, proto, ui_client: bool = True, share: bool = False,
                       uuid="") -> tuple[bool, int]:
        share_count = 0
        disconnected = 0
        existing_sources = set(ss for p, ss in self._server_sources.items() if p != proto)
        is_existing_client = uuid and any(ss.uuid == uuid for ss in existing_sources)
        authlog("handle_sharing%s lock=%s, sharing=%s, existing sources=%s, is existing client=%s",
                (proto, ui_client, share, uuid),
                self.lock, self.sharing, existing_sources, is_existing_client)
        # if other clients are connected, verify we can steal or share:
        if existing_sources and not is_existing_client:
            if self.sharing is True or (self.sharing is None and share and all(ss.share for ss in existing_sources)):
                authlog("handle_sharing: sharing with %s", tuple(existing_sources))
            elif self.lock is True:
                authlog("handle_sharing: session is locked")
                self.disconnect_client(proto, ConnectionMessage.SESSION_BUSY, "this session is locked")
                return False, 0
            elif self.lock is not False and any(ss.lock for ss in existing_sources):
                authlog("handle_sharing: another client has locked the session: " + csv(
                    ss for ss in existing_sources if ss.lock))
                self.disconnect_client(proto, ConnectionMessage.SESSION_BUSY, "a client has locked this session")
                return False, 0
        # we're either sharing, or the only client:
        for p, ss in tuple(self._server_sources.items()):
            if uuid and ss.uuid == uuid and ui_client and ss.ui_client:
                authlog("uuid %s is the same as %s", uuid, ss)
                authlog("existing sources: %s", existing_sources)
                self.disconnect_client(p, ConnectionMessage.NEW_CLIENT, "new connection from the same uuid")
                disconnected += 1
            elif ui_client and ss.ui_client:
                # check if existing sessions are willing to share:
                if self.sharing is True:
                    share_count += 1
                elif self.sharing is False:
                    self.disconnect_client(p, ConnectionMessage.NEW_CLIENT, "this session does not allow sharing")
                    disconnected += 1
                else:
                    assert self.sharing is None
                    if not share:
                        self.disconnect_client(p, ConnectionMessage.NEW_CLIENT, "the new client does not wish to share")
                        disconnected += 1
                    elif not ss.share:
                        self.disconnect_client(p, ConnectionMessage.NEW_CLIENT, "this client had not enabled sharing")
                        disconnected += 1
                    else:
                        share_count += 1

        # don't accept this connection if we're going to exit-with-client:
        accepted = True
        if disconnected > 0 and share_count == 0 and self.exit_with_client:
            self.disconnect_client(proto, ConnectionMessage.SERVER_SHUTDOWN, "last client has exited")
            accepted = False
        return accepted, share_count

    def hello_oked(self, proto, c: typedict, auth_caps: dict) -> None:
        if self._server_sources.get(proto):
            log.warn("Warning: received another 'hello' packet")
            log.warn(" from an existing connection: %s", proto)
            return
        if ServerCore.hello_oked(self, proto, c, auth_caps):
            # has been handled
            return
        if not self.sanity_checks(proto, c):
            return
        if not c.boolget("steal", True) and self._server_sources:
            self.disconnect_client(proto, ConnectionMessage.SESSION_BUSY, "this session is already active")
            return
        log.info("Handshake complete; enabling connection")
        self.server_event("handshake-complete")

        # Things are okay, we accept this connection, and may disconnect previous one(s)
        # (but only if this is going to be a UI session - control sessions can co-exist)
        ui_client = c.boolget("ui_client", True)
        share = c.boolget("share")
        uuid = c.strget("uuid")
        accepted, share_count = self.handle_sharing(proto, ui_client, share, uuid)
        if not accepted:
            return

        self.accept_client(proto, c)
        # use blocking sockets from now on:
        if not WIN32:
            set_socket_timeout(proto._conn, None)

        def drop_client(reason="unknown", *args) -> None:
            self.disconnect_client(proto, reason, *args)

        cc_class = self.get_client_connection_class(c)
        ss = cc_class(proto, drop_client, self, self.setting_changed)
        log("process_hello clientconnection=%s", ss)
        try:
            ss.parse_hello(c)
        except Exception:
            # close it already
            ss.close()
            raise
        self._server_sources[proto] = ss
        add_work_item(self.mdns_update)
        # process ui half in ui thread:
        GLib.idle_add(self.process_hello_ui, ss, c, auth_caps, ui_client, share_count)

    def do_handle_hello_request(self, request: str, proto, caps: typedict) -> bool:
        if super().do_handle_hello_request(request, proto, caps):
            return True
        if request == "detach":
            self.detach_server(proto)
            return True
        if request == "exit":
            self._process_exit_server(proto)
            return True
        if request == "stop":
            self._process_shutdown_server(proto)
            return True
        return False

    def detach_server(self, proto) -> None:
        if self.lock is True:
            authlog("cannot detach: session is locked")
            self.disconnect_client(proto, ConnectionMessage.SESSION_BUSY, "this session is locked")
            return
        count = locked = 0
        for p, ss in tuple(self._server_sources.items()):
            if p != proto:
                if ss.lock:
                    locked += 1
                else:
                    authlog("handle_sharing: detaching %s", ss)
                    self.disconnect_client(p, ConnectionMessage.DETACH_REQUEST)
                    count += 1
        message = f"{count} clients have been disconnected"
        if locked:
            message += f", {locked} still have it locked"
        self.disconnect_client(proto, ConnectionMessage.DONE, message)

    @staticmethod
    def get_client_connection_class(caps: typedict) -> type:
        # pylint: disable=import-outside-toplevel
        from xpra.server.source.client_connection_factory import get_client_connection_class
        return get_client_connection_class(caps)

    def process_hello_ui(self, ss, c: typedict, auth_caps: dict, send_ui: bool, share_count: int) -> None:
        def reject(message="server is shutting down") -> None:
            p = ss.protocol
            if p:
                self.disconnect_client(p, ConnectionMessage.CONNECTION_ERROR, message)

        # adds try:except around parse hello ui code:
        try:
            if self._closing:
                reject()
                return

            self.notify_new_user(ss)

            self.parse_hello(ss, c, send_ui)
            # send_hello will take care of sending the current and max screen resolutions

            self.send_hello(ss, auth_caps)
            self.add_new_client(ss, c, send_ui, share_count)
            self.send_initial_data(ss, c, send_ui, share_count)
            self.client_startup_complete(ss)

            if self._closing:
                reject()
                return
        except Exception:
            # log exception but don't disclose internal details to the client
            log("process_hello_ui%s", (ss, c, auth_caps, send_ui, share_count))
            log.error("Error: processing new connection from %s:", ss.protocol or ss, exc_info=True)
            reject("error accepting new connection")

    def parse_hello(self, ss, c: typedict, send_ui: bool) -> None:
        for bc in SERVER_BASES:
            if bc != ServerCore:
                bc.parse_hello(self, ss, c, send_ui)

    def add_new_client(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        for bc in SERVER_BASES:
            if bc != ServerCore:
                bc.add_new_client(self, ss, c, send_ui, share_count)

    def send_initial_data(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        for bc in SERVER_BASES:
            if bc != ServerCore:
                bc.send_initial_data(self, ss, c, send_ui, share_count)

    def client_startup_complete(self, ss) -> None:
        ss.startup_complete()
        self.server_event("startup-complete", ss.uuid)

    def sanity_checks(self, proto, c: typedict) -> bool:
        server_uuid = c.strget("server_uuid")
        if server_uuid:
            if server_uuid == self.uuid:
                self.send_disconnect(proto, "cannot connect a client running on the same display"
                                            " that the server it connects to is managing - this would create a loop!")
                return False
            log.warn("Warning: this client is running nested")
            log.warn(f" in the Xpra server session {server_uuid!r}")
        return True

    def update_all_server_settings(self, reset: bool = False) -> None:
        pass  # may be overridden in subclasses (ie: x11 server)

    ######################################################################
    # hello:
    def get_server_features(self, server_source=None) -> dict[str, Any]:
        # these are flags that have been added over time with new versions
        # to expose new server features:
        f: dict[str, Any] = {}
        for c in SERVER_BASES:
            if c != ServerCore:
                bf = c.get_server_features(self, server_source)
                log(f"get_server_features({c})={bf}")
                merge_dicts(f, bf)
        return f

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = super().make_hello(source)
        for c in SERVER_BASES:
            if c != ServerCore:
                caps = c.get_caps(self, source)
                log("%s.get_caps(%s)=%s", c, source, caps)
                merge_dicts(capabilities, caps)
        capabilities["server_type"] = "base"
        if "features" in source.wants:
            capabilities |= {
                "client-shutdown": self.client_shutdown,
                "sharing": self.sharing is not False,
                "sharing-toggle": self.sharing is None,
                "lock": self.lock is not False,
                "lock-toggle": self.lock is None,
                "windows": features.windows,
                "keyboard": features.input_devices,
            }
            sf = self.get_server_features(source)
            capabilities.update(sf)
        return capabilities

    def send_hello(self, server_source, server_cipher: dict) -> None:
        capabilities = self.make_hello(server_source)
        if server_cipher:
            capabilities.update(server_cipher)
        server_source.send_hello(capabilities)
        self.after_threaded_init(server_source.threaded_init_complete, self)

    ######################################################################
    # utility method:
    def window_sources(self) -> tuple:
        from xpra.server.source.windows import WindowsMixin  # pylint: disable=import-outside-toplevel
        return tuple(x for x in self._server_sources.values() if isinstance(x, WindowsMixin))

    ######################################################################
    # info:
    def _process_info_request(self, proto, packet: PacketType) -> None:
        log("process_info_request(%s, %s)", proto, packet)
        # ignoring the list of client uuids supplied in packet[1]
        ss = self.get_server_source(proto)
        if not ss:
            return

        try:
            options = proto._conn.options
            info_option = options.get("info", "yes")
        except AttributeError:
            info_option = "yes"
        if not str_to_bool(info_option):
            err = "`info` commands are not enabled on this connection"
            log.warn(f"Warning: {err}")
            ss.send_info_response({"error": err})
            return

        categories: list[str] = []
        # if len(packet>=2):
        #    uuid = packet[1]
        if len(packet) >= 4:
            categories = [str(x) for x in packet[3]]

        def info_callback(_proto, info) -> None:
            assert proto == _proto
            if categories:
                info = {k: v for k, v in info.items() if k in categories}
            ss.send_info_response(info)

        self.get_all_info(info_callback, proto, None)

    def send_hello_info(self, proto) -> None:
        self.wait_for_threaded_init()
        start = monotonic()

        def cb(iproto, info) -> None:
            self.do_send_info(iproto, info)
            end = monotonic()
            log.info("processed info request from %s in %ims",
                     iproto._conn, (end - start) * 1000)

        self.get_all_info(cb, proto, None)

    def get_ui_info(self, proto, client_uuids=None, *args) -> dict[str, Any]:
        """ info that must be collected from the UI thread
            (ie: things that query the display)
        """
        info: dict[str, Any] = {}
        for c in SERVER_BASES:
            with log.trap_error("Error collecting UI info from %s", c):
                mixin_info = c.get_ui_info(self, proto, client_uuids, *args)
                log("%s.get_ui_info(%s, ..)=%r", c, proto, Ellipsizer(mixin_info))
                merge_dicts(info, mixin_info)
        return info

    def get_info(self, proto=None, client_uuids=None) -> dict[str, Any]:
        log("ServerBase.get_info%s", (proto, client_uuids))
        start = monotonic()
        if client_uuids:
            sources = tuple(ss for ss in self._server_sources.values() if ss.uuid in client_uuids)
        else:
            sources = tuple(self._server_sources.values())
        log("info-request: sources=%s", sources)
        info = self.do_get_info(proto, sources)
        log("ServerBase.get_info took %.1fms", 1000.0 * (monotonic() - start))
        return info

    def get_features_info(self) -> dict[str, Any]:
        i = {
            "sharing": self.sharing is not False,
            "idle_timeout": self.idle_timeout,
        }
        i.update(self.get_server_features())
        return i

    def do_get_info(self, proto, server_sources=()) -> dict[str, Any]:
        log("ServerBase.do_get_info%s", (proto, server_sources))
        start = monotonic()
        info: dict[str, Any] = {}

        def up(prefix, d) -> None:
            merge_dicts(info, {prefix: d})

        subsystems = []
        for c in SERVER_BASES:
            subsystems.append(c.__name__.replace("Server", ""))
            with log.trap_error(f"Error collecting information from {c}"):
                cstart = monotonic()
                mixin_info = c.get_info(self, proto)
                log("%s.get_info(%s)=%r", c, proto, Ellipsizer(mixin_info))
                merge_dicts(info, mixin_info)
                cend = monotonic()
                log("%s.get_info(%s) took %ims", c, proto, int(1000 * (cend - cstart)))
        up("subsystems", subsystems)

        up("features", self.get_features_info())
        up("network", {
            "sharing": self.sharing is not False,
            "sharing-toggle": self.sharing is None,
            "lock": self.lock is not False,
            "lock-toggle": self.lock is None,
        })

        # other clients:
        info["clients"] = {
            "": sum(1 for p in self._server_sources if p != proto),
            "unauthenticated": sum(1 for p in self._potential_protocols
                                   if ((p is not proto) and (p not in self._server_sources))),
        }
        log("unauthenticated protocols:")
        for i, p in enumerate(self._potential_protocols):
            log(f"{i:3} : {p}={p.get_info()}")
        # find the server source to report on:
        n = len(server_sources)
        if n == 1:
            ss = server_sources[0]
            up("client", ss.get_info())
        elif n > 1:
            cinfo = {}
            for i, ss in enumerate(server_sources):
                sinfo = ss.get_info()
                sinfo["ui-driver"] = self.ui_driver == ss.uuid
                cinfo[i] = sinfo
            up("client", cinfo)
        log("ServerBase.do_get_info took %ims", (monotonic() - start) * 1000)
        return info

    def _process_server_settings(self, proto, packet: PacketType) -> None:
        # only used by x11 servers
        pass

    def _set_client_properties(self, proto, wid: int, window, new_client_properties: dict) -> None:
        """
        Allows us to keep window properties for a client after disconnection.
        (we keep it in a map with the client's uuid as key)
        """
        ss = self.get_server_source(proto)
        if ss:
            ss.set_client_properties(wid, window, typedict(new_client_properties))
            # filter out encoding properties, which are expected to be set every time:
            ncp = {}
            for k, v in new_client_properties.items():
                if v is None:
                    log.warn("removing invalid None property for %s", k)
                    continue
                k = str(k)
                if k == "event":
                    # event is used as a workaround in _process_map_window,
                    # it isn't a real client property and should not be stored:
                    continue
                if not k.startswith("encoding"):
                    ncp[k] = v
            if ncp:
                log("set_client_properties updating window %s of source %s with %s", wid, ss.uuid, ncp)
                client_properties = self.client_properties.setdefault(wid, {}).setdefault(ss.uuid, {})
                client_properties.update(ncp)

    ######################################################################
    # settings toggle:
    def setting_changed(self, setting: str, value: Any) -> None:
        # tell all the clients (that can) about the new value for this setting
        for ss in tuple(self._server_sources.values()):
            ss.send_setting_change(setting, value)

    def _process_sharing_toggle(self, proto, packet: PacketType) -> None:
        assert self.sharing is None
        ss = self.get_server_source(proto)
        if not ss:
            return
        sharing = bool(packet[1])
        ss.share = sharing
        if not sharing:
            # disconnect other users:
            for p, ss in tuple(self._server_sources.items()):
                if p != proto:
                    self.disconnect_client(p, ConnectionMessage.DETACH_REQUEST,
                                           f"client {ss.counter} no longer wishes to share the session")

    def _process_lock_toggle(self, proto, packet: PacketType) -> None:
        assert self.lock is None
        ss = self.get_server_source(proto)
        if ss:
            ss.lock = bool(packet[1])
            log("lock set to %s for client %i", ss.lock, ss.counter)

    ######################################################################
    # add clients to http server info:
    def get_http_info(self) -> dict[str, Any]:
        info = super().get_http_info()
        info["clients"] = len(self._server_sources)
        return info

    ######################################################################
    # client connections:
    def init_sockets(self, sockets) -> None:
        for c in SERVER_BASES:
            c.init_sockets(self, sockets)

    def cleanup_protocol(self, protocol):
        netlog("cleanup_protocol(%s)", protocol)
        # this ensures that from now on we ignore any incoming packets coming
        # from this connection as these could potentially set some keys pressed, etc
        try:
            self._potential_protocols.remove(protocol)
        except ValueError:
            pass
        source = self._server_sources.pop(protocol, None)
        if source:
            self.cleanup_source(source)
            add_work_item(self.mdns_update)
        for c in SERVER_BASES:
            c.cleanup_protocol(self, protocol)
        return source

    def cleanup_source(self, source) -> None:
        ptype = "xpra"
        if FULL_INFO > 0:
            ptype = getattr(source, "client_type", "") or "xpra"
        self.server_event("connection-lost", source.uuid)
        remaining_sources = tuple(self._server_sources.values())
        if self.ui_driver == source.uuid:
            if len(remaining_sources) == 1:
                self.set_ui_driver(remaining_sources[0])
            else:
                self.set_ui_driver(None)
        source.close()
        netlog("cleanup_source(%s) remaining sources: %s", source, remaining_sources)
        netlog.info("%s client %i disconnected.", ptype, source.counter)
        has_client = len(remaining_sources) > 0
        if not has_client:
            GLib.idle_add(self.last_client_exited)

    def last_client_exited(self) -> None:
        # must run from the UI thread (modifies focus and keys)
        netlog("last_client_exited() exit_with_client=%s", self.exit_with_client)
        self.reset_server_timeout(True)
        for c in SERVER_BASES:
            if c != ServerCore:
                try:
                    c.last_client_exited(self)
                except Exception:
                    log("last_client_exited calling %s", c.last_client_exited, exc_info=True)
        if self.exit_with_client and not self._closing:
            netlog.info("Last client has disconnected, terminating")
            self.clean_quit(False)

    def set_ui_driver(self, source) -> None:
        if source and self.ui_driver == source.uuid:
            return
        log("new ui driver: %s", source)
        if not source:
            self.ui_driver = None
        else:
            self.ui_driver = source.uuid
        for c in SERVER_BASES:
            if c != ServerCore:
                c.set_session_driver(self, source)

    def get_all_protocols(self) -> list:
        return list(self._potential_protocols) + list(self._server_sources.keys())

    def is_timedout(self, protocol) -> bool:
        v = ServerCore.is_timedout(self, protocol) and protocol not in self._server_sources
        netlog("is_timedout(%s)=%s", protocol, v)
        return v

    def _disconnect_proto_info(self, proto) -> str:
        # only log protocol info if there is more than one client:
        if len(self._server_sources) > 1:
            return " %s" % proto
        return ""

    def init_packet_handlers(self) -> None:
        for c in SERVER_BASES:
            c.init_packet_handlers(self)
        # no need for main thread:
        self.add_packets("sharing-toggle", "lock-toggle")
        self.add_packet_handler("set_deflate", noop)  # removed in v6
        # attributes / settings:
        self.add_packets("server-settings", "info-request", main_thread=True)
        self.add_packets("shutdown-server", "exit-server")

    # override so we can set the 'authenticated' flag:
    def process_packet(self, proto, packet) -> None:
        authenticated = bool(self.get_server_source(proto))
        return super().dispatch_packet(proto, packet, authenticated)

    def handle_invalid_packet(self, proto, packet) -> None:
        ss = self.get_server_source(proto)
        if not self._closing and not proto.is_closed() and (ss is None or not ss.is_closed()):
            netlog("invalid packet: %s", packet)
            packet_type = str(packet[0])
            netlog.error(f"Error: unknown or invalid packet type {packet_type!r}")
            netlog.error(f" received from {proto}")
        if not ss:
            proto.close()
